# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class TrainingStatusExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, training_guid=None, training_url=None, training_reference_name=None, training_definition_url=None, submitted_at=None, finished_at=None, current_at=None, current_iteration=None, total_iterations=None, state=None, message=None, metrics=None, hyper_parameters=None, result=None, failure=None):
        """
        TrainingStatusExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'training_guid': 'str',
            'training_url': 'str',
            'training_reference_name': 'str',
            'training_definition_url': 'str',
            'submitted_at': 'datetime',
            'finished_at': 'datetime',
            'current_at': 'datetime',
            'current_iteration': 'float',
            'total_iterations': 'float',
            'state': 'str',
            'message': 'str',
            'metrics': 'ArrayModelVersionMetricsExperiments',
            'hyper_parameters': 'HyperParametersForStatusExperiments',
            'result': 'TrainingStatusExperimentsResult',
            'failure': 'ErrorSchemaExperiments'
        }

        self.attribute_map = {
            'training_guid': 'training_guid',
            'training_url': 'training_url',
            'training_reference_name': 'training_reference_name',
            'training_definition_url': 'training_definition_url',
            'submitted_at': 'submitted_at',
            'finished_at': 'finished_at',
            'current_at': 'current_at',
            'current_iteration': 'current_iteration',
            'total_iterations': 'total_iterations',
            'state': 'state',
            'message': 'message',
            'metrics': 'metrics',
            'hyper_parameters': 'hyper_parameters',
            'result': 'result',
            'failure': 'failure'
        }

        self._training_guid = training_guid
        self._training_url = training_url
        self._training_reference_name = training_reference_name
        self._training_definition_url = training_definition_url
        self._submitted_at = submitted_at
        self._finished_at = finished_at
        self._current_at = current_at
        self._current_iteration = current_iteration
        self._total_iterations = total_iterations
        self._state = state
        self._message = message
        self._metrics = metrics
        self._hyper_parameters = hyper_parameters
        self._result = result
        self._failure = failure

    @property
    def training_guid(self):
        """
        Gets the training_guid of this TrainingStatusExperiments.


        :return: The training_guid of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._training_guid

    @training_guid.setter
    def training_guid(self, training_guid):
        """
        Sets the training_guid of this TrainingStatusExperiments.


        :param training_guid: The training_guid of this TrainingStatusExperiments.
        :type: str
        """

        self._training_guid = training_guid

    @property
    def training_url(self):
        """
        Gets the training_url of this TrainingStatusExperiments.
        Currently, this is link to current /V3 training API, so to the /v3/models.

        :return: The training_url of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._training_url

    @training_url.setter
    def training_url(self, training_url):
        """
        Sets the training_url of this TrainingStatusExperiments.
        Currently, this is link to current /V3 training API, so to the /v3/models.

        :param training_url: The training_url of this TrainingStatusExperiments.
        :type: str
        """

        self._training_url = training_url

    @property
    def training_reference_name(self):
        """
        Gets the training_reference_name of this TrainingStatusExperiments.


        :return: The training_reference_name of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._training_reference_name

    @training_reference_name.setter
    def training_reference_name(self, training_reference_name):
        """
        Sets the training_reference_name of this TrainingStatusExperiments.


        :param training_reference_name: The training_reference_name of this TrainingStatusExperiments.
        :type: str
        """

        self._training_reference_name = training_reference_name

    @property
    def training_definition_url(self):
        """
        Gets the training_definition_url of this TrainingStatusExperiments.


        :return: The training_definition_url of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._training_definition_url

    @training_definition_url.setter
    def training_definition_url(self, training_definition_url):
        """
        Sets the training_definition_url of this TrainingStatusExperiments.


        :param training_definition_url: The training_definition_url of this TrainingStatusExperiments.
        :type: str
        """

        self._training_definition_url = training_definition_url

    @property
    def submitted_at(self):
        """
        Gets the submitted_at of this TrainingStatusExperiments.


        :return: The submitted_at of this TrainingStatusExperiments.
        :rtype: datetime
        """
        return self._submitted_at

    @submitted_at.setter
    def submitted_at(self, submitted_at):
        """
        Sets the submitted_at of this TrainingStatusExperiments.


        :param submitted_at: The submitted_at of this TrainingStatusExperiments.
        :type: datetime
        """

        self._submitted_at = submitted_at

    @property
    def finished_at(self):
        """
        Gets the finished_at of this TrainingStatusExperiments.


        :return: The finished_at of this TrainingStatusExperiments.
        :rtype: datetime
        """
        return self._finished_at

    @finished_at.setter
    def finished_at(self, finished_at):
        """
        Sets the finished_at of this TrainingStatusExperiments.


        :param finished_at: The finished_at of this TrainingStatusExperiments.
        :type: datetime
        """

        self._finished_at = finished_at

    @property
    def current_at(self):
        """
        Gets the current_at of this TrainingStatusExperiments.


        :return: The current_at of this TrainingStatusExperiments.
        :rtype: datetime
        """
        return self._current_at

    @current_at.setter
    def current_at(self, current_at):
        """
        Sets the current_at of this TrainingStatusExperiments.


        :param current_at: The current_at of this TrainingStatusExperiments.
        :type: datetime
        """

        self._current_at = current_at

    @property
    def current_iteration(self):
        """
        Gets the current_iteration of this TrainingStatusExperiments.


        :return: The current_iteration of this TrainingStatusExperiments.
        :rtype: float
        """
        return self._current_iteration

    @current_iteration.setter
    def current_iteration(self, current_iteration):
        """
        Sets the current_iteration of this TrainingStatusExperiments.


        :param current_iteration: The current_iteration of this TrainingStatusExperiments.
        :type: float
        """

        self._current_iteration = current_iteration

    @property
    def total_iterations(self):
        """
        Gets the total_iterations of this TrainingStatusExperiments.


        :return: The total_iterations of this TrainingStatusExperiments.
        :rtype: float
        """
        return self._total_iterations

    @total_iterations.setter
    def total_iterations(self, total_iterations):
        """
        Sets the total_iterations of this TrainingStatusExperiments.


        :param total_iterations: The total_iterations of this TrainingStatusExperiments.
        :type: float
        """

        self._total_iterations = total_iterations

    @property
    def state(self):
        """
        Gets the state of this TrainingStatusExperiments.


        :return: The state of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """
        Sets the state of this TrainingStatusExperiments.


        :param state: The state of this TrainingStatusExperiments.
        :type: str
        """
        allowed_values = ["running", "completed", "failed", "canceled"]
        if state not in allowed_values:
            raise ValueError(
                "Invalid value for `state` ({0}), must be one of {1}"
                .format(state, allowed_values)
            )

        self._state = state

    @property
    def message(self):
        """
        Gets the message of this TrainingStatusExperiments.


        :return: The message of this TrainingStatusExperiments.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this TrainingStatusExperiments.


        :param message: The message of this TrainingStatusExperiments.
        :type: str
        """

        self._message = message

    @property
    def metrics(self):
        """
        Gets the metrics of this TrainingStatusExperiments.


        :return: The metrics of this TrainingStatusExperiments.
        :rtype: ArrayModelVersionMetricsExperiments
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """
        Sets the metrics of this TrainingStatusExperiments.


        :param metrics: The metrics of this TrainingStatusExperiments.
        :type: ArrayModelVersionMetricsExperiments
        """

        self._metrics = metrics

    @property
    def hyper_parameters(self):
        """
        Gets the hyper_parameters of this TrainingStatusExperiments.


        :return: The hyper_parameters of this TrainingStatusExperiments.
        :rtype: HyperParametersForStatusExperiments
        """
        return self._hyper_parameters

    @hyper_parameters.setter
    def hyper_parameters(self, hyper_parameters):
        """
        Sets the hyper_parameters of this TrainingStatusExperiments.


        :param hyper_parameters: The hyper_parameters of this TrainingStatusExperiments.
        :type: HyperParametersForStatusExperiments
        """

        self._hyper_parameters = hyper_parameters

    @property
    def result(self):
        """
        Gets the result of this TrainingStatusExperiments.


        :return: The result of this TrainingStatusExperiments.
        :rtype: TrainingStatusExperimentsResult
        """
        return self._result

    @result.setter
    def result(self, result):
        """
        Sets the result of this TrainingStatusExperiments.


        :param result: The result of this TrainingStatusExperiments.
        :type: TrainingStatusExperimentsResult
        """

        self._result = result

    @property
    def failure(self):
        """
        Gets the failure of this TrainingStatusExperiments.


        :return: The failure of this TrainingStatusExperiments.
        :rtype: ErrorSchemaExperiments
        """
        return self._failure

    @failure.setter
    def failure(self, failure):
        """
        Sets the failure of this TrainingStatusExperiments.


        :param failure: The failure of this TrainingStatusExperiments.
        :type: ErrorSchemaExperiments
        """

        self._failure = failure

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import os
import logging
import logging.config
import time

from ibmfl._version import __version__
from ibmfl.exceptions import InvalidConfigurationException

logger = logging.getLogger(__name__)


class UTCFormatter(logging.Formatter):
    converter = time.gmtime

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            if "%F" in datefmt:
                msec = "%03d" % record.msecs
                datefmt = datefmt.replace("%F", msec)
            s = time.strftime(datefmt, ct)
        else:
            t = time.strftime("%Y-%m-%dT%H:%M:%S.%F", ct)
            s = "%s.%03dZ" % (t, record.msecs)
        return s


class FLVersionFilter(logging.Filter):

    def filter(self, record):
        record.version = __version__
        return True


def configure_file_logging(filename, log_level='INFO'):
    log_level = logging.getLevelName(log_level)

    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "fl_std": {
                "()": UTCFormatter,
                "format": "%(asctime)s | %(version)s | %(levelname)s | %(name)-50s | %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S.%FZ"
            }
        },
        "handlers": {
            "WriteFile": {
                "class": "logging.FileHandler",
                "filters": ['version_filter'],
                "level": "DEBUG",
                "formatter": "fl_std",
                "filename": filename
            }

        },
        "loggers": {
            "ibmfl": {
                "level": log_level,
                "handlers": [
                    "WriteFile"
                ],
                "propagate": False
            }
        }
    }

    result = configure_logging(dict_config, log_level)
    return result


def configure_logging(config, log_level=logging.INFO):
    """
    configures logging for application based on the configuration file

    :param config: yaml file containing the definitions of formatter and handler
    :type config: `dict`
    :param log_level: should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]
            based on the required granularity
    :type log_level: `int`
    :return: a boolean object. False for default basic config True otherwise
    :rtype: `boolean`
    """

    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "fl_std": {
                "()": UTCFormatter,
                "format": "%(asctime)s | %(version)s | %(levelname)s | %(name)-50s | %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S.%FZ"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "filters": ['version_filter'],
                "level": log_level,
                "formatter": "fl_std",
                "stream": "ext://sys.stdout"
            }

        },
        "loggers": {
            "ibmfl": {
                "level": log_level,
                "handlers": [
                    "console"
                ],
                "propagate": False
            }
        },
        "root": {
            "level": log_level,
            "handlers": [
                "console"
            ]
        }
    }
    if not config:
        config = dict_config

    log_filters = {
        'version_filter': {
            '()': FLVersionFilter
        }
    }
    if 'filters' in config:
        config['filters']['version_filter'] = log_filters['version_filter']
    else:
        config['filters'] = log_filters

    add_version_filter(config)
    result = False
    if config:
        try:
            logging.config.dictConfig(config)
            result = True
        except InvalidConfigurationException as ice:
            logging.error('Failed to load log configuration %s', ice)
            configure_basic_logging(log_level=log_level)
    else:
        configure_basic_logging(log_level=log_level)

    return result


def add_version_filter(config):
    """
    Add versioning filter to log config if not provided

    :param config: yaml file containing the definitions of formatter and handler
    :type config: `dict`
    :return: None
    """
    filters = ['version_filter']
    if 'handlers' in config:

        handlers = config['handlers']

        for key in handlers:
            handler = handlers.get(key)

            if 'filters' in handler:
                filt = handler.get('filters')
                if 'version_filter' not in filt:
                    filt.append('version_filter')
            else:
                handler['filters'] = filters


def configure_basic_logging(log_level=logging.INFO):
    """
    configures logging for the session based on pre defined format.
    default logging will be done on console.

    :param level: should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]
            based on the required granularity
    :type log_level: `int`
    :return: None
    """

    logging.basicConfig(level=log_level,
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # logging.config.dictConfig(default_log_config)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2019- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.metanames import ModelDefinitionMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError
import os
import json
import uuid

_DEFAULT_LIST_LENGTH = 50


class ModelDefinition(WMLResource):
    """Store and manage model definitions."""

    ConfigurationMetaNames = ModelDefinitionMetaNames()

    """MetaNames for model definition creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP
        self.default_space_id = client.default_space_id

    def _generate_model_definition_document(self, meta_props):
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            doc = {
                "metadata":
                {
                   "name": "generated_name_"+str(uuid.uuid4()),
                   "tags": ["generated_tag_"+str(uuid.uuid4())],
                   "asset_type": "wml_model_definition",
                   "origin_country": "us",
                   "rov": {
                      "mode": 0
                   },
                   "asset_category": "USER"
                },
                "entity": {
                   "wml_model_definition": {
                       "ml_version": "4.0.0",
                       "version": "1.0",
                       "platform": {
                          "name": "python",
                          "versions": [
                              "3.5"
                             ]
                       }
                   }
                }
            }

        else:
            doc = {
                "metadata":
                {
                   "name": "My wml_model_definition assert",
                   "tags": ["string"],
                   "asset_type": "wml_model_definition",
                   "origin_country": "us",
                   "rov": {
                      "mode": 0
                   },
                   "asset_category": "USER"
                },
                "entity": {
                   "wml_model_definition": {
                       "name": "tf-model_trainings_v4_test_suite_basic",
                       "description": "Sample custom library",
                       "version": "1.0",
                       "platform": {
                          "name": "python",
                          "versions": [
                              "3.5"
                             ]
                       }
                   }
                }
            }

        if self.ConfigurationMetaNames.NAME in meta_props:
            doc["metadata"]["name"] = meta_props[self.ConfigurationMetaNames.NAME]
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                # We shouldn't have name and description in entity but this code exists in CP4D
                # So, changing only for Cloud Convergence
                doc["entity"]["wml_model_definition"]["name"] = meta_props[self.ConfigurationMetaNames.NAME]
        if self.ConfigurationMetaNames.DESCRIPTION in meta_props:
            doc["metadata"]["description"] = meta_props[self.ConfigurationMetaNames.DESCRIPTION]
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                doc["entity"]["wml_model_definition"]["description"] = meta_props[self.ConfigurationMetaNames.DESCRIPTION]

        if self.ConfigurationMetaNames.VERSION in meta_props:
            doc["entity"]["wml_model_definition"]["version"] = meta_props[self.ConfigurationMetaNames.VERSION]

        if self.ConfigurationMetaNames.PLATFORM in meta_props:
            doc["entity"]["wml_model_definition"]["platform"]["name"] = meta_props[self.ConfigurationMetaNames.PLATFORM]['name']
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                doc["entity"]["wml_model_definition"]["platform"]["versions"] = \
                    meta_props[self.ConfigurationMetaNames.PLATFORM]['versions']
            else:
                doc["entity"]["wml_model_definition"]["platform"]["versions"][0] = \
                    meta_props[self.ConfigurationMetaNames.PLATFORM]['versions'][0]

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            if self.ConfigurationMetaNames.COMMAND in meta_props:
                doc['entity']['wml_model_definition'].update({"command": meta_props[self.ConfigurationMetaNames.COMMAND]})
            if self.ConfigurationMetaNames.CUSTOM in meta_props:
                doc['entity']['wml_model_definition'].update(
                    {"custom": meta_props[self.ConfigurationMetaNames.CUSTOM]})

        return doc

    def store(self, model_definition, meta_props):
        """Create a model definition.

        :param meta_props: meta data of the model definition configuration, to see available meta names use:

            .. code-block:: python

                client.model_definitions.ConfigurationMetaNames.get()

        :type meta_props: dict

        :param model_definition: path to the content file to be uploaded
        :type model_definition: str

        :return: metadata of the model definition created
        :rtype: dict

        **Example**

        .. code-block:: python

            client.model_definitions.store(model_definition, meta_props)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        self.ConfigurationMetaNames._validate(meta_props)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:

            # metadata doesn't seem to be used at all.. Code existing pre-convergence.
            # Should be checked sometime and removed if not needed
            metadata = {
                self.ConfigurationMetaNames.NAME: meta_props[self.ConfigurationMetaNames.NAME],
                self.ConfigurationMetaNames.VERSION: meta_props['version'],
                self.ConfigurationMetaNames.PLATFORM:
                    meta_props['platform']
                    if 'platform' in meta_props and meta_props['platform'] is not None
                    else {
                        "name": meta_props[self.ConfigurationMetaNames.PLATFORM]['name'],
                        "versions": [meta_props[self.ConfigurationMetaNames.PLATFORM]['versions']]
                    },
            }

            if self.ConfigurationMetaNames.DESCRIPTION in meta_props:
                metadata[self.ConfigurationMetaNames.DESCRIPTION] = meta_props[self.ConfigurationMetaNames.DESCRIPTION]

            if self.ConfigurationMetaNames.SPACE_UID in meta_props:
                metadata[self.ConfigurationMetaNames.SPACE_UID] = {
                    "href": "/v4/spaces/" + meta_props[self.ConfigurationMetaNames.SPACE_UID]
                }

            # Following is not used for model_definitions since space_id/project_id
            # is passed as a query param. Code is existing pre-cloud-convergence
            if self._client.CAMS:
                if self._client.default_space_id is not None:
                    metadata['space'] = {'href': "/v4/spaces/"+self._client.default_space_id}
                elif self._client.default_project_id is not None:
                    metadata['project'] = {'href': "/v2/projects/"+self._client.default_project_id}
                else:
                    raise WMLClientError("It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")

        document = self._generate_model_definition_document(meta_props)

        if self._client.WSD:
            return self._wsd_create_asset("wml_model_definition", document, meta_props, model_definition,user_archive_file=True)

        model_definition_attachment_def = {
          "asset_type": "wml_model_definition",
          "name": "model_definition_attachment"
        }
        paramvalue = self._client._params()

        creation_response = requests.post(
             self._client.service_instance._href_definitions.get_model_definition_assets_href(),
             params=paramvalue,
             headers=self._client._get_headers(),
             json=document)

        model_definition_details = self._handle_response(201, u'creating new model_definition', creation_response)

        self._handle_response(201, 'creating new attachment', creation_response)
        model_definition_id = model_definition_details['metadata']['asset_id']

        if self._client.CLOUD_PLATFORM_SPACES:
            model_definition_attachment_url = self._client.service_instance._href_definitions.get_attachments_href(model_definition_details['metadata']['asset_id'])
        else:
            model_definition_attachment_url = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + \
                                              model_definition_details['metadata']['asset_id'] + "/attachments"

        put_header = self._client._get_headers(no_content_type=True)

        attachment_response = requests.post(
            model_definition_attachment_url,
            params=paramvalue,
            headers=self._client._get_headers(),
            json=model_definition_attachment_def)

        try:
            attachment_details = self._handle_response(201, u'creating new model definition attachment', attachment_response)

            attachment_id = attachment_details['attachment_id']
            attachment_status_json = json.loads(attachment_response.content.decode("utf-8"))
            model_definition_attachment_signed_url = attachment_status_json["url1"]
         #   print("WML model_definition attachment url1: %s" % model_definition_attachement_signed_url)
            model_definition_attachment_put_url = self._client.wml_credentials['url'] + model_definition_attachment_signed_url

            with open(model_definition, 'rb') as f:
                if self._ICP:
                    put_response = requests.put(model_definition_attachment_put_url,
                                            files={'file': (model_definition, f, 'application/octet-stream')})
                else:
                    put_response = requests.put(model_definition_attachment_signed_url,
                                                data=open(model_definition, 'rb').read())

            if put_response.status_code != 201 and put_response.status_code != 200:
                self._handle_response(200, 'uploading a model_definition attachment file', put_response)

            complete_response = requests.post( self._client.service_instance._href_definitions.get_attachment_complete_href(model_definition_id, attachment_id),
                                              params=paramvalue,
                                              headers=self._client._get_headers())

            self._handle_response(200, 'updating a model_definition status', complete_response)

            response = self._get_required_element_from_response(model_definition_details)

            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:
                entity = response[u'entity']

                try:
                    del entity[u'wml_model_definition'][u'ml_version']
                except KeyError:
                    pass

                final_response = {
                    "metadata": response[u'metadata'],
                    "entity": entity
                }

                return final_response
            # return self._get_required_element_from_response(model_definition_details)
        except Exception as e:
            try:
                self._delete(model_definition_id)
            finally:
                raise e

    def get_details(self, model_definition_uid=None):
        """Get metadata of stored model definition. If no `model_definition_uid` is passed,
        details for all model definitions will be returned.

        :param model_definition_uid: Unique Id of model definition
        :type model_definition_uid: str, optional

        :return: metadata of model definition
        :rtype: dict (if `model_definition_uid` is not None)

        **Example**

        .. code-block: python

            model_definition_details = client.model_definitions.get_details(model_definition_uid)

        """
        def get_required_element_from_response(response):
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:
                final_response = {
                    "metadata": response[u'metadata'],
                }

                if 'entity' in response:
                    entity = response[u'entity']

                    try:
                        del entity[u'wml_model_definition'][u'ml_version']
                    except KeyError:
                        pass

                    final_response["entity"] = entity

                return final_response

        return self._get_asset_based_resource(model_definition_uid, 'wml_model_definition', get_required_element_from_response)

    def download(self, model_definition_uid, filename, rev_id=None):
        """Download the content of a model definition asset.

        :param model_definition_uid: the Unique Id of the model definition asset to be downloaded
        :type model_definition_uid: str
        :param filename: filename to be used for the downloaded file
        :type filename: str
        :param rev_id: revision id
        :type rev_id: str, optional

        :return: path to the downloaded asset content
        :rtype: str

        **Example**

        .. code-block:: python

            client.model_definitions.download(model_definition_uid, "model_definition_file")
        """

        self._client._check_if_either_is_set()

        ModelDefinition._validate_type(model_definition_uid, u'model_definition_uid', str, True)
        params = self._client._params()
        if rev_id is not None:
            ModelDefinition._validate_type(rev_id, u'rev_id', int, False)
            params.update({'revision_id': rev_id})

        if self._client.WSD:
            import urllib
            response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(model_definition_uid),
                                              params=self._client._params(),
                                              headers=self._client._get_headers())

            model_def_details = self._handle_response(200, u'get model', response)
            attachment_url = model_def_details['attachments'][0]['object_key']
            attachment_signed_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                   urllib.parse.quote('wml_model_definition/' + attachment_url, safe='')

        else:
            attachment_id = self._get_attachment_id(model_definition_uid)
            artifact_content_url = self._client.service_instance._href_definitions.get_attachment_href(model_definition_uid, attachment_id)
            if not self._ICP and not self._WSD:
                response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(model_definition_uid, attachment_id), params=self._client._params(),
                                    headers=self._client._get_headers())
            else:
                response = requests.get(artifact_content_url, params=self._client._params(),
                                    headers=self._client._get_headers())
            attachment_signed_url = response.json()["url"]
        if response.status_code == 200:
            if not self._ICP and not self._client.WSD:
                if self._client.CLOUD_PLATFORM_SPACES:
                    att_response = requests.get(attachment_signed_url)
                else:
                    att_response = requests.get(self._wml_credentials["url"] + attachment_signed_url)
            else:
                if self._client.WSD:
                    att_response = requests.get(attachment_signed_url, params= self._client._params(),
                                                headers=self._client._get_headers(),
                                                stream=True)
                else:
                    att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url)

            if att_response.status_code != 200:
                raise WMLClientError(u'Failure during {}.'.format("downloading model_definition asset"),
                                     att_response)

            downloaded_asset = att_response.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved asset content to file: \'{}\''.format(filename))
                return os.getcwd() + "/" + filename
            except IOError as e:
                raise WMLClientError(u'Saving asset with artifact_url: \'{}\' failed.'.format(filename), e)
        else:
            raise WMLClientError("Failed while downloading the asset " + model_definition_uid)

    def delete(self, model_definition_uid):
        """Delete a stored model definition.

        :param model_definition_uid: Unique Id of stored model definition
        :type model_definition_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.model_definitions.delete(model_definition_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        ModelDefinition._validate_type(model_definition_uid, u'model_definition_uid', str, True)
        paramvalue = self._client._params()

        model_definition_endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + model_definition_uid

        response_delete = requests.delete(model_definition_endpoint, params=paramvalue, headers=self._client._get_headers())

        return self._handle_response(204, u'Model definition deletion', response_delete, False)

    def _delete(self, model_definition_uid):
        """Delete a stored model definition.

        :param model_definition_uid: Unique Id of stored model definition
        :type model_definition_uid: str

        **Example**

        .. code-block:: python

            client.model_definitions._delete(model_definition_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        ModelDefinition._validate_type(model_definition_uid, u'model_definition_uid', str, True)
        paramvalue = self._client._params()
        model_definition_endpoint = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + model_definition_uid

        response_delete = requests.delete(model_definition_endpoint, params=paramvalue, headers=self._client._get_headers())

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data,  u'model_definition_response', dict)
        revision_id = None

        try:
            href = ""
            if self._client.default_space_id is not None:
                new_el = {'metadata': {'space_id': response_data['metadata']['space_id'],
                                   'guid': response_data['metadata']['asset_id'],
                                   'asset_type': response_data['metadata']['asset_type'],
                                   'created_at': response_data['metadata']['created_at'],
                                   'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                },
                      'entity': response_data['entity']

                      }
                href = self._client.service_instance._href_definitions.get_base_asset_with_type_href(response_data['metadata']['asset_type'], response_data['metadata'][
                    'asset_id']) + "?" + "space_id=" + response_data['metadata']['space_id']

            elif self._client.default_project_id is not None:
                new_el = {'metadata': {'project_id': response_data['metadata']['project_id'],
                                       'guid': response_data['metadata']['asset_id'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at'],
                                       'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                       },
                          'entity': response_data['entity']

                          }

                href = self._client.service_instance._href_definitions.get_base_asset_with_type_href(response_data['metadata']['asset_type'], response_data['metadata'][
                    'asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

            if 'revision_id' in response_data['metadata']:
                new_el['metadata'].update({'revision_id': response_data['metadata']['revision_id']})
                revision_id = response_data[u'metadata'][u'revision_id']

            if 'name' in response_data['metadata']:
                new_el['metadata'].update({'name': response_data['metadata']['name']})

            if 'description' in response_data['metadata'] and response_data['metadata']['description']:
                new_el['metadata'].update({'description': response_data['metadata']['description']})

            if 'href' in response_data['metadata']:
                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    href_without_host = response_data['href'].split('.com')[-1]
                    new_el[u'metadata'].update({'href': href_without_host})
                else:
                    new_el['metadata'].update({'href': response_data['href']})
            else:
                new_el['metadata'].update({'href': href})

            if "attachments" in response_data and response_data[u'attachments']:
                new_el['metadata'].update({'attachment_id': response_data[u'attachments'][0][u'id']})
            else:
                new_el['metadata'].update({'href': href})

            if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                new_el['metadata'].update(
                    {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})
            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data)

    def _get_attachment_id(self, model_definition_uid, rev_id=None):
        op_name = 'getting attachment id '
        url = self._client.service_instance._href_definitions.get_model_definition_assets_href() + u'/' + model_definition_uid
        paramvalue = self._client._params()

        if rev_id is not None:
            paramvalue.update({'revision_id': rev_id})

        response_get = requests.get(
            url,
            params=paramvalue,
            headers=self._client._get_headers()
        )
        details = self._handle_response(200, op_name, response_get)
        try:
            attachment_id = details["attachments"][0]["id"]
        except KeyError:
            raise WMLClientError(f'No attachment exists for model definition (id={model_definition_uid}).')

        return attachment_id

    def list(self, limit=None):
        """Print stored model definition assets in a table format.
        If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.model_definitions.list()

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        href = self._client.service_instance._href_definitions.get_model_definition_search_asset_href()
        if limit is None:
            data = {
                "query": "*:*"
            }
        else:
            ModelDefinition._validate_type(limit, u'limit', int, False)
            data = {
                "query": "*:*",
                "limit": limit
            }

        response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), json=data)

        self._handle_response(200, u'model_definition assets', response)
        asset_details = self._handle_response(200, u'model_definition assets', response)["results"]
        model_def_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m[u'metadata'][u'asset_id']) for
            m in asset_details]

        self._list(model_def_values, [u'NAME', u'ASSET_TYPE', u'GUID'], limit, _DEFAULT_LIST_LENGTH)

    def get_id(self, model_definition_details):
        """Get Unique Id of stored model definition asset.

        :param model_definition_details: metadata of the stored model definition asset
        :type model_definition_details: dict

        :return: Unique Id of stored model definition asset
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.model_definition.get_id(asset_details)

        """

        return ModelDefinition.get_uid(self, model_definition_details)

    def get_uid(self, model_definition_details):
        """Get uid of stored model.

        *Deprecated:* Use ``get_id(model_definition_details)`` instead.

        :param model_definition_details: stored model definition details
        :type model_definition_details: dict

        :return: uid of stored model definition
        :rtype: str

        **Example**

        .. code-block:: python

            model_definition_uid = client.model_definitions.get_uid(model_definition_details)
        """
        if 'asset_id' in model_definition_details['metadata']:
            return WMLResource._get_required_element_from_dict(model_definition_details, u'model_definition_details',
                                                               [u'metadata', u'asset_id'])
        else:
            ModelDefinition._validate_type(model_definition_details, u'model__definition_details', object, True)
            #ModelDefinition._validate_type_of_details(model_definition_details, MODEL_DEFINITION_DETAILS_TYPE)

            return WMLResource._get_required_element_from_dict(model_definition_details, u'model_definition_details', [u'metadata', u'guid'])

    def get_href(self, model_definition_details):
        """Get href of stored model definition.

        :param model_definition_details: stored model definition details
        :type model_definition_details: dict

        :return: href of stored model definition
        :rtype: str

        **Example**

        .. code-block:: python

            model_definition_uid = client.model_definitions.get_href(model_definition_details)
        """
        if 'asset_id' in model_definition_details['metadata']:
            return WMLResource._get_required_element_from_dict(model_definition_details, u'model_definition_details', [u'metadata', u'asset_id'])
        else:
            ModelDefinition._validate_type(model_definition_details, u'model__definition_details', object, True)
            # ModelDefinition._validate_type_of_details(model_definition_details, MODEL_DEFINITION_DETAILS_TYPE)

            return WMLResource._get_required_element_from_dict(model_definition_details, u'model_definition_details',
                                                               [u'metadata', u'href'])

    def update(self, model_definition_id, meta_props=None, file_path=None):
        """Update model definition with either metadata or attachment or both.

        :param model_definition_id: model definition ID
        :type model_definition_id: str
        :param meta_props: meta data of the model definition configuration to be updated
        :type meta_props: dict
        :param file_path: path to the content file to be uploaded
        :type file_path: str, optional

        :return: updated metadata of model definition
        :rtype: dict

        **Example**

        .. code-block:: python

            model_definition_details = client.model_definition.update(model_definition_id, meta_props, file_path)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        # We need to enable this once we add functionality for WSD
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ModelDefinition._validate_type(model_definition_id, 'model_definition_id', str, True)

        if meta_props is None and file_path is None:
            raise WMLClientError('At least either meta_props or file_path has to be provided')

        updated_details = None

        url = self._client.service_instance._href_definitions.get_asset_href(model_definition_id)

        # STEPS
        # STEP 1. Get existing metadata
        # STEP 2. If meta_props provided, we need to patch meta
        #   CAMS has meta and entity patching. 'name' and 'description' get stored in CAMS meta section
        #   a. Construct meta patch string and call /v2/assets/<asset_id> to patch meta
        #   b. Construct entity patch if required and call /v2/assets/<asset_id>/attributes/script to patch entity
        # STEP 3. If file_path provided, we need to patch the attachment
        #   a. If attachment already exists for the model_definition, delete it
        #   b. POST call to get signed url for upload
        #   c. Upload to the signed url
        #   d. Mark upload complete
        # STEP 4. Get the updated script record and return

        # STEP 1
        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of model_definition_id provided.')
            else:
                raise WMLClientError(u'Failure during {}.'.format("getting script to update"), response)

        details = self._handle_response(200, "Get script details", response)

        attachments_response = None

        # STEP 2a.
        # Patch meta if provided
        if meta_props is not None:
            self._validate_type(meta_props, u'meta_props', dict, True)

            props_for_asset_meta_patch = {}

            # Since we are dealing with direct asset apis, there can be metadata or entity patch or both
            if "name" in meta_props or "description" in meta_props:

                for key in meta_props:
                    if key == 'name' or key == 'description':
                        props_for_asset_meta_patch.update({key: meta_props[key]})

            meta_patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details,
                                                                                     props_for_asset_meta_patch,
                                                                                     with_validation=True)

            props_for_asset_entity_patch = {}
            for key in meta_props:
                if key != 'name' and key != 'description':
                    props_for_asset_entity_patch.update({key: meta_props[key]})

            entity_patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity']['wml_model_definition'],
                                                                                       props_for_asset_entity_patch,
                                                                                       with_validation=True)

            if meta_patch_payload:
                meta_patch_url = self._client.service_instance._href_definitions.get_asset_href(model_definition_id)

                response_patch = requests.patch(meta_patch_url,
                                                json=meta_patch_payload,
                                                params=self._client._params(),
                                                headers=self._client._get_headers())

                updated_details = self._handle_response(200, u'script patch', response_patch)

            if entity_patch_payload:
                entity_patch_url = self._client.service_instance._href_definitions.get_asset_href(model_definition_id) + \
                                   '/attributes/wml_model_definition'

                response_patch = requests.patch(entity_patch_url,
                                                json=entity_patch_payload,
                                                params=self._client._params(),
                                                headers=self._client._get_headers())

                updated_details = self._handle_response(200, u'script patch', response_patch)

        if file_path is not None:
            if "attachments" in details and details[u'attachments']:
                current_attachment_id = details[u'attachments'][0][u'id']
            else:
                current_attachment_id = None

            #STEP 3
            attachments_response = self._update_attachment_for_assets("wml_model_definition",
                                                                      model_definition_id,
                                                                      file_path,
                                                                      current_attachment_id)

        if attachments_response is not None and 'success' not in attachments_response:
            self._update_msg(updated_details)

        # Have to fetch again to reflect updated asset and attachment ids
        url = self._client.service_instance._href_definitions.get_asset_href(model_definition_id)

        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of model_definition_id provided.')
            else:
                raise WMLClientError(u'Failure during {}.'.format("getting script to update"), response)

        response = self._get_required_element_from_response(self._handle_response(200, "Get script details", response))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'wml_model_definition'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

        # return self._get_required_element_from_response(response)

    def _update_msg(self, updated_details):
        if updated_details is not None:
            print("Could not update the attachment because of server error."
                  " However metadata is updated. Try updating attachment again later")
        else:
            raise WMLClientError('Unable to update attachment because of server error. Try again later')

    def create_revision(self, model_definition_uid):
        """Create revision for the given model definition. Revisions are immutable once created.
        The metadata and attachment at model definition is taken and a revision is created out of it.

        :param model_definition_uid: model definition ID
        :type model_definition_uid: str

        :return: stored model definition revisions metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            model_definition_revision = client.model_definitions.create_revision(model_definition_id)
        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                u'Revisions APIs are not supported in this release.')

        self._client._check_if_either_is_set()

        ModelDefinition._validate_type(model_definition_uid, u'model_definition_uid', str, True)

        print("Creating model_definition revision...")

        # return self._get_required_element_from_response(
        #     self._create_revision_artifact_for_assets(model_defn_id, 'Model definition'))

        response = self._get_required_element_from_response(
            self._create_revision_artifact_for_assets(model_definition_uid, 'Model definition'))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'wml_model_definition'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def get_revision_details(self, model_definition_uid, rev_uid=None):
        """Get metadata of model definition.

        :param model_definition_uid: model definition ID
        :type model_definition_uid: str

        :param rev_uid: revision ID, if this parameter is not provided, returns latest revision if existing else error
        :type rev_uid: int, optional

        :return: stored model definitions metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            script_details = client.model_definitions.get_revision_details(model_definition_uid, rev_uid)
        """

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                u'Revisions APIs are not supported in this release.')
        op_name = 'getting model_definition revision details'
        ModelDefinition._validate_type(model_definition_uid, u'model_definition_uid', str, True)

        url = self._client.service_instance._href_definitions.get_model_definition_assets_href() + u'/' + model_definition_uid
        paramvalue = self._client._params()

        if rev_uid is None:
            rev_uid = 'latest'

        paramvalue.update({'revision_id': rev_uid})

        response_get = requests.get(
            url,
            params=paramvalue,
            headers=self._client._get_headers()
        )
        if response_get.status_code == 200:
            # get_model_definition_details = self._handle_response(200, op_name, response_get)
            # return self._get_required_element_from_response(get_model_definition_details)

            response = self._get_required_element_from_response(self._handle_response(200, op_name, response_get))

            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:
                entity = response[u'entity']

                try:
                    del entity[u'wml_model_definition'][u'ml_version']
                except KeyError:
                    pass

                final_response = {
                    "metadata": response[u'metadata'],
                    "entity": entity
                }

                return final_response
        else:
            return self._handle_response(200, op_name, response_get)

    def list_revisions(self, model_definition_uid, limit=None):
        """Print stored model definition assets in a table format.
        If limit is set to None there will be only first 50 records shown.

        :param model_definition_uid: Unique id of model definition
        :type model_definition_uid: str
        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.model_definitions.list_revisions()

        """
        ##For CP4D, check if either spce or project ID is set
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                u'Revisions APIs are not supported in this release.')
        self._client._check_if_either_is_set()
        href = self._client.service_instance._href_definitions.get_model_definition_assets_href() + "/" + model_definition_uid +\
               u'/revisions'
        params = self._client._params()
        #params = None
        if limit is not None:
            ModelDefinition._validate_type(limit, u'limit', int, False)
            params.update( {
                "limit": limit
            })
        response = requests.get(href, params=params, headers=self._client._get_headers())
        self._handle_response(200, u'model_definition revision assets', response)
        asset_details = self._handle_response(200, u'model_definition revision assets', response)["results"]
        model_def_values = [
            (m[u'metadata'][u'asset_id'],
             m[u'metadata'][u'revision_id'],
             m[u'metadata'][u'name'],
             m[u'metadata'][u'asset_type'],
             m[u'metadata'][u'commit_info'][u'committed_at']) for
            m in asset_details]

        self._list(model_def_values, [u'GUID', u'REV_ID', u'NAME', u'ASSET_TYPE', u'REVISION_COMMIT'],
                   limit,
                   _DEFAULT_LIST_LENGTH)






#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.helpers.helpers import *
from ibm_watson_machine_learning.helpers.connections import (
    DataConnection, S3Connection, S3Location, FSLocation, AssetLocation, CP4DAssetLocation, WMLSAssetLocation,
    WSDAssetLocation, NFSLocation, NFSConnection, ConnectionAsset, ConnectionAssetLocation, DatabaseLocation,
    ContainerLocation)



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import abc
from enum import Enum


class FLConnection(abc.ABC):

    SENDER_STATUS = None
    RECEIVER_STATUS = None

    """
    Initialze sender and receiver

    :param kwargs: Dictionary of arguments required by each type of
        sender implementation
    """
    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Validate config dictionary and initialize receiver

    :param kwargs: Dictionary of arguments required by each type of
        receiver implementation
    """
    @abc.abstractmethod
    def initialize_receiver(self, **kwargs):
        pass

    """
    Validate config dictionary and initialize sender

    :param kwargs: Dictionary of arguments required by each type of
        sender implementation
    """
    @abc.abstractmethod
    def initialize_sender(self, **kwargs):
        pass

    """
    Start connection
    """
    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    """Stop and cleanup connection object
    """
    @abc.abstractmethod
    def stop(self, **kwargs):
        pass

    """Provide a connection information such that a node can communicate 
        to other nodes on how to communicate with it.
    """
    @abc.abstractmethod
    def get_connection_config(self):
        pass


class FLReceiver(abc.ABC):

    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Start the server as per configuration after a clean intialize_server
    call
    :param kwargs: Dictionary of arguments required by each type of
        server implementation
    """
    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    """
    Cleanup all the resources and stop serving
    :param kwargs: Dictionary of arguments required by each type of
        server implementation
    """
    @abc.abstractmethod
    def stop(self, **kwargs):
        pass


class FLSender(abc.ABC):

    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Requests the end point as per the message forwarded to this module.
    Should have a logic to incorporate additional authentication details as
    per the configuration.

    :param kwargs: Dictionary of arguments required by each type of
        connection implementation
    """
    @abc.abstractmethod
    def send_message(self, **kwargs):
        pass

    """
    Cleanup the Sender object and close the hanging connections if any
    """
    @abc.abstractmethod
    def cleanup(self):
        pass


class ConnectionStatus(Enum):
    """Enum Class to describe the current status of the connection
    """
    INITIALIZED = 1
    STARTED = 2
    STOPPED = 3



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import abc
import logging

from ibmfl.data.data_handler import DataHandler

logger = logging.getLogger(__name__)


class PandasDataHandler(DataHandler):
    """
    Base class to load and pre-process data.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.x_train = None
        self.y_train = None
        self.x_test = None
        self.y_test = None

    @abc.abstractmethod
    def get_data(self):
        """
        Read data and return as Pandas data frame.

        :return: A dataset structure
        :rtype: `pandas.core.frame.DataFrame`
        """

    @abc.abstractmethod
    def get_dataset_info(self, **kwargs):
        """
        Read and extract data information

        :return: some information about the dataset (i.e. a dictionary that contains the list of features)
        :rtype: `dict`
        """
        raise NotImplemented

    def get_min(self, dp_flag=False, **kwargs):
        """
        Assuming the dataset is loaded as type `pandas.DataFrame`, and
        has shape(num_samples, num_features).

        :param dp_flag: Flag for differential private answer. By default is \
        set to False.
        :type dp_flag: `boolean`
        :param kwargs: Dictionary of differential privacy arguments \
        for computing the minimum value of each feature across all samples, \
        e.g., epsilon and delta, etc.
        :type kwargs: `dict`
        :return: A vector of shape (1, num_features) stores the minimum value \
        of each feature across all samples.
        :rtype: `pandas.Series` where each entry matches the original type \
        of the corresponding feature.
        """
        train_data, (_) = self.get_data()

        if not dp_flag:
            logger.info('Calculating minimum values.')
            min_vec = train_data.min(axis=0)
            # TODO dp minimum calculation
        return min_vec



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2019- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
import json
from ibm_watson_machine_learning.utils import SPACES_IMPORTS_DETAILS_TYPE, SPACES_EXPORTS_DETAILS_TYPE, SPACES_DETAILS_TYPE, MEMBER_DETAILS_TYPE, print_text_header_h2
from ibm_watson_machine_learning.metanames import SpacesMetaNames, MemberMetaNames, ExportMetaNames
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import  WMLClientError
from ibm_watson_machine_learning.utils.deployment.errors import PromotionFailed

_DEFAULT_LIST_LENGTH = 50


class Spaces(WMLResource):
    """Store and manage spaces. This is applicable only for IBM Cloud Pak® for Data."""

    ConfigurationMetaNames = SpacesMetaNames()
    """MetaNames for spaces creation."""
    MemberMetaNames = MemberMetaNames()
    """MetaNames for space members creation."""
    ExportMetaNames = ExportMetaNames()
    """MetaNames for spaces export."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)

        self._ICP = client.ICP

    def store(self, meta_props):
        """Create a space.

        :param meta_props: meta data of the space configuration. To see available meta names use:

            .. code-block:: python

                client.spaces.ConfigurationMetaNames.get()

        :type meta_props: dict

        :return: metadata of the stored space
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.ConfigurationMetaNames.NAME: 'my_space',
                client.spaces.ConfigurationMetaNames.DESCRIPTION: 'spaces',
            }
            spaces_details = client.spaces.store(meta_props=metadata)
        """

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        WMLResource._chk_and_block_create_update_for_python36(self)
        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        creation_response = requests.post(
            self._wml_credentials['url'] + '/v4/spaces',
            headers=self._client._get_headers(),
            json=space_meta
        )

        spaces_details = self._handle_response(201, u'creating new spaces', creation_response)

        # Cloud Convergence: Set self._client.wml_credentials['instance_id'] to instance_id
        # during client.set.default_space since that's where space is associated with client
        # and also in client.set.default_project

        return spaces_details

    @staticmethod
    def get_href(spaces_details):
        """Get space href from space details.

        :param spaces_details: metadata of the stored space
        :type spaces_details: dict

        :return: space href
        :rtype: str

        **Example**

        .. code-block:: python

            space_details = client.spaces.get_details(space_uid)
            space_href = client.spaces.get_href(space_details)
        """

        Spaces._validate_type(spaces_details, u'spaces_details', object, True)
        Spaces._validate_type_of_details(spaces_details, SPACES_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(spaces_details, u'spaces_details',
                                                           [u'metadata', u'href'])

    @staticmethod
    def get_uid(spaces_details):
        """Get space uid from space details.

        :param spaces_details: metadata of the stored space
        :type spaces_details: dict

        :return: space UID
        :rtype: str

        **Example**

        .. code-block:: python

            space_details = client.spaces.get_details(space_uid)
            space_uid = client.spaces.get_uid(space_details)
        """

        Spaces._validate_type(spaces_details, u'spaces_details', object, True)
        Spaces._validate_type_of_details(spaces_details, SPACES_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(spaces_details, u'spaces_details',
                                                           [u'metadata', u'guid'])

    def delete(self, space_uid):
        """Delete a stored space.

        :param space_uid: space UID
        :type space_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.spaces.delete(sapce_uid)
        """
        Spaces._validate_type(space_uid, u'space_uid', str, True)

        space_endpoint = self._client.service_instance._href_definitions.get_space_href(space_uid)

        response_delete = requests.delete(space_endpoint, headers=self._client._get_headers())

        return self._handle_response(204, u'space deletion', response_delete, False)

    def get_details(self, space_uid=None, limit=None):
        """Get metadata of stored space(s). If space UID is not specified, it returns all the spaces metadata.

        :param space_uid: Space UID
        :type space_uid: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional

        :return: metadata of stored space(s)
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)

        **Example**

        .. code-block:: python

            space_details = client.spaces.get_details(space_uid)
            space_details = client.spaces.get_details()
        """
        Spaces._validate_type(space_uid, u'space_uid', str, False)
        Spaces._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.get_spaces_href()
        if space_uid is None:
            return self._get_no_space_artifact_details(href+"?include=name,tags,custom,description", None, limit, 'spaces')
        return self._get_no_space_artifact_details(href, space_uid, limit, 'spaces')

    def list(self, limit=None):
        """Print stored spaces in a table format. If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.spaces.list()
        """

        space_resources = self.get_details(limit=limit)[u'resources']
        space_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at']) for m in space_resources]

        self._list(space_values, [u'GUID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

    def update(self, space_uid, changes):
        """Updates existing space metadata.

        :param space_uid: UID of space which definition should be updated
        :type space_uid: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated space
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.ConfigurationMetaNames.NAME: "updated_space"
            }
            space_details = client.spaces.update(space_uid, changes=metadata)
        """
        self._validate_type(space_uid, u'space_uid', str, True)
        self._validate_type(changes, u'changes', dict, True)

        details = self.get_details(space_uid)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes,
                                                                            with_validation=True)

        href = self._client.service_instance._href_definitions.get_space_href(space_uid)

        response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'spaces patch', response)

        return updated_details


#######SUPPORT FOR SPACE MEMBERS

    ###GET MEMBERS DETAILS

    def get_members_details(self, space_uid, member_id=None, limit=None):
        """Get metadata of members associated with a space.
        If member UID is not specified, it returns all the members metadata.

        :param space_uid: space UID
        :type space_uid: str
        :param member_id: member UID
        :type member_id: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional

        :return: metadata of member(s) of a space
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)

        **Example**

        .. code-block:: python

            member_details = client.spaces.get_members_details(space_uid,member_id)
        """
        Spaces._validate_type(space_uid, u'space_uid', str, True)
        Spaces._validate_type(member_id, u'member_id', str, False)
        Spaces._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.get_members_href(space_uid)

        return self._get_no_space_artifact_details(href, member_id, limit, 'space members')

    ##DELETE MEMBERS

    def delete_members(self, space_uid, member_id):
        """Delete a member associated with a space.

        :param space_uid: space UID
        :type space_uid: str
        :param member_id: member UID
        :type member_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.spaces.delete_member(space_uid,member_id)
        """
        Spaces._validate_type(space_uid, u'space_uid', str, True)
        Spaces._validate_type(member_id, u'member_id', str, True)

        member_endpoint = self._client.service_instance._href_definitions.get_member_href(space_uid,member_id)

        response_delete = requests.delete(member_endpoint, headers=self._client._get_headers())

        return self._handle_response(204, u'space member deletion', response_delete, False)

#######UPDATE MEMBERS

    def update_member(self, space_uid, member_id, changes):
        """Updates existing member metadata.

        :param space_uid: UID of space
        :type space_uid: str
        :param member_id: UID of member that needs to be updated
        :type member_id: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated member
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.ConfigurationMetaNames.ROLE:"viewer"
            }
            member_details = client.spaces.update_member(space_uid, member_id, changes=metadata)
        """
        self._validate_type(space_uid, u'space_uid', str, True)
        self._validate_type(member_id, u'member_id', str, True)

        self._validate_type(changes, u'changes', dict, True)

        details = self.get_members_details(space_uid,member_id)

        patch_payload = self.MemberMetaNames._generate_patch_payload(details['entity'], changes,
                                                                            with_validation=True)

        href = self._client.service_instance._href_definitions.get_member_href(space_uid, member_id)

        response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'members patch', response)

        return updated_details

#####CREATE MEMBER
    def create_member(self, space_uid, meta_props):
        """Create a member within a space.

        :param space_uid: UID of space
        :type space_uid: str
        :param meta_props:  metadata of the member configuration. To see available meta names use:

            .. code-block:: python

                client.spaces.MemberMetaNames.get()

        :type meta_props: dict

        :return: metadata of the stored member
        :rtype: dict

        .. note::
            * client.spaces.MemberMetaNames.ROLE can be any one of the following "viewer", "editor", "admin"
            * client.spaces.MemberMetaNames.IDENTITY_TYPE can be any one of the following "user", "service"
            * client.spaces.MemberMetaNames.IDENTITY can be either service-ID or IAM-userID

        **Example**

        .. code-block:: python

            metadata = {
                client.spaces.MemberMetaNames.ROLE:"Admin",
                client.spaces.MemberMetaNames.IDENTITY:"iam-ServiceId-5a216e59-6592-43b9-8669-625d341aca71",
                client.spaces.MemberMetaNames.IDENTITY_TYPE:"service"
            }
            members_details = client.spaces.create_member(space_uid=space_id, meta_props=metadata)
        """

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_meta = self.MemberMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        creation_response = requests.post(
            self._wml_credentials['url'] + '/v4/spaces/'+space_uid+"/members",
            headers=self._client._get_headers(),
            json=space_meta
        )

        members_details = self._handle_response(201, u'creating new members', creation_response)

        return members_details

    def list_members(self, space_uid, limit=None):
        """Print stored members of a space in a table format.
        If limit is set to None there will be only first 50 records shown.

        :param space_uid: UID of space
        :type space_uid: str
        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.spaces.list_members()
        """

        member_resources = self.get_members_details(space_uid,limit=limit)[u'resources']
        space_values = [(m[u'metadata'][u'guid'],  m[u'entity'][u'identity'], m[u'entity'][u'identity_type'], m[u'entity'][u'role'], m[u'metadata'][u'created_at']) for m in member_resources]

        self._list(space_values, [u'GUID', u'USERNAME', u'IDENTITY_TYPE', u'ROLE', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

    @staticmethod
    def get_member_href(member_details):
        """Get member href from member details.

        :param member_details: metadata of the stored member
        :type member_details: dict

        :return: member href
        :rtype: str

        **Example**

        .. code-block:: python

            member_details = client.spaces.get_member_details(member_id)
            member_href = client.spaces.get_member_href(member_details)
        """

        Spaces._validate_type(member_details, u'member details', object, True)
        Spaces._validate_type_of_details(member_details, MEMBER_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(member_details, u'member_details',
                                                           [u'metadata', u'href'])

    @staticmethod
    def get_member_uid(member_details):
        """Get member uid from member details.

        :param member_details: metadata of the created member
        :type member_details: dict

        :return: member UID
        :rtype: str

        **Example**

        .. code-block:: python

            member_details = client.spaces.get_member_details(member_id)
            member_id = client.spaces.get_member_uid(member_details)
        """

        Spaces._validate_type(member_details, u'member_details', object, True)
        Spaces._validate_type_of_details(member_details, MEMBER_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(member_details, u'member_details',
                                                           [u'metadata', u'guid'])

    def imports(self, space_uid, file_path):
        """Imports assets in the zip file to a space. Updates existing space metadata.

        :param space_uid: UID of space which definition should be updated
        :type space_uid: str
        :param file_path: path to the content file to be imported
        :type file_path: str

        :return: metadata of import space
        :rtype: dict

        **Example**

        .. code-block:: python

            space_details = client.spaces.imports(space_uid, file_path="/tmp/spaces.zip")
        """
        self._validate_type(space_uid, u'space_uid', str, True)
        self._validate_type(file_path, u'file_path', str, True)



        with open(str(file_path), 'rb') as archive:
            data = archive.read()

        href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports"

        response = requests.post(href, headers=self._client._get_headers(), data=data)

        import_space_details = self._handle_response(202, u'spaces import', response)

        return import_space_details



    @staticmethod
    def get_imports_uid(imports_space_details):
        """Get imports_uid from imports space details.

        :param imports_space_details: metadata of the created space import
        :type imports_space_details: dict

        :return: imports space UID
        :rtype: str

        **Example**

        .. code-block:: python

            imports_space_details = client.spaces.get_imports_details(space_uid, imports_id)
            imports_id = client.spaces.get_imports_uid(imports_space_details)
        """

        Spaces._validate_type(imports_space_details, u'member_details', object, True)
        Spaces._validate_type_of_details(imports_space_details, SPACES_IMPORTS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(imports_space_details, u'imports_space_details',[u'metadata', u'guid'])

    @staticmethod
    def get_exports_uid(exports_space_details):
        """Get imports_uid from imports space details.

        :param exports_space_details: metadata of the created space import
        :type exports_space_details: dict

        :return: exports space UID
        :rtype: str

        **Example**

        .. code-block:: python

            exports_space_details = client.spaces.get_exports_details(space_uid, exports_id)
            exports_id = client.spaces.get_exports_uid(exports_space_details)
        """

        Spaces._validate_type(exports_space_details, u'exports_space_details', object, True)
        Spaces._validate_type_of_details(exports_space_details, SPACES_EXPORTS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(exports_space_details, u'exports_space_details',
                                                           [u'metadata', u'guid'])

    def get_imports_details(self, space_uid, imports_id=None, limit=None):
        """Get metadata of space import(s). If imports UID is not specified, it returns all the spaces imports details.

        :param space_uid: space UID
        :type space_uid: str
        :param imports_id: imports UID
        :type imports_id: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional

        :return: metadata of space import(s)
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)

        **Examples**

        .. code-block:: python

            space_details = client.spaces.get_imports_details(space_uid)
            space_details = client.spaces.get_imports_details(space_uid, imports_id)
        """
        Spaces._validate_type(space_uid, u'space_uid', str, False)
        Spaces._validate_type(imports_id, u'imports_uid', str, False)
        Spaces._validate_type(limit, u'limit', int, False)


        if imports_id is None:
            href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports"
        else:
            href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports/" + imports_id

        response = requests.get(href, headers=self._client._get_headers())

        import_space_details = self._handle_response(200, u'spaces import details', response)

        return import_space_details

    def exports(self, space_uid, meta_props):
        """Exports assets in the zip file from a space. Updates existing space metadata.

        :param space_uid: space UID
        :type space_uid: str
        :param meta_props: meta data of the space configuration. To see available meta names use:

            .. code-block:: python

                client.spaces.ExportMetaNames.get()

        :type meta_props: dict

        :return: metadata of exports space
        :rtype: dict

        **Example**

        .. code-block:: python

            meta_props = {
                client.spaces.ExportMetaNames.NAME: "sample",
                client.spaces.ExportMetaNames.DESCRIPTION : "test description",
                client.spaces.ExportMetaNames.ASSETS : {"data_assets": [], "wml_model":[]} }
            }
            space_details = client.spaces.exports(space_uid, meta_props=meta_props)

        """

        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_exports_meta = self.ExportMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        space_exports_meta_json = json.dumps(space_exports_meta)

        self._validate_type(space_uid, u'space_uid', str, True)

        href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports"

        response = requests.post(href, headers=self._client._get_headers(), data=space_exports_meta_json)

        export_space_details = self._handle_response(202, u'spaces export', response)

        return export_space_details

    def get_exports_details(self, space_uid, exports_id=None, limit=None):
        """Get details of exports for space. If exports UID is not specified, it returns all the space exports metadata.

        :param space_uid: space UID
        :type space_uid: str
        :param exports_id: exports UID
        :type exports_id: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional

        :return: metadata of exports of space
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n

        **Example**

        .. code-block:: python

            space_details = client.spaces.get_exports_details(space_uid)
            space_details = client.spaces.get_exports_details(space_uid, exports_id)

        """
        Spaces._validate_type(space_uid, u'space_uid', str, False)
        Spaces._validate_type(exports_id, u'imports_uid', str, False)
        Spaces._validate_type(limit, u'limit', int, False)


        if exports_id is None:
            href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports"
        else:
            href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports/" + exports_id

        response = requests.get(href, headers=self._client._get_headers())

        export_space_details = self._handle_response(200, u'spaces exports details', response)

        return export_space_details

    def download(self, space_uid, space_exports_uid, filename=None):
        """Downloads zip file deployment of specified UID.

        :param space_uid: space UID
        :type space_uid: str
        :param space_exports_uid: UID of virtual deployment
        :type space_exports_uid: str
        :param filename: filename of downloaded archive
        :type filename: str, optional

        :return: path to downloaded file
        :rtype: str

        **Example**

        .. code-block:: python

            client.spaces.download(space_uid)
        """
        Spaces._validate_type(space_exports_uid, u'space_exports_uid', str, False)

        if space_exports_uid is not None and not is_uid(space_exports_uid):
            raise WMLClientError(u'\'space_exports_uid\' is not an uid: \'{}\''.format(space_exports_uid))

        href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports/" + space_exports_uid + "/content"

        response = requests.get(
            href,
            headers=self._client._get_headers()
        )

        if filename is None:
            filename = 'wmlspace.zip'

        if response.status_code == 200:
            with open(filename, "wb") as new_file:
                new_file.write(response.content)
                new_file.close()

                print_text_header_h2(
                    u'Successfully downloaded spaces export file: ' + str(filename))

                return filename
        else:
            raise WMLClientError(u'Unable to download spaces export: ' + response.text)

    def promote(self, asset_id: str, source_project_id: str, target_space_id: str, rev_id: str = None) -> str:
        """Promote asset from project to space.

        :param asset_id: stored asset
        :type asset_id: str

        :param source_project_id: source project, from which asset is promoted
        :type source_project_id: str

        :param target_space_id: target space, where asset is promoted
        :type target_space_id: str

        :param rev_id: revision ID of the promoted asset
        :type rev_id: str, optional

        :return: promoted asset id
        :rtype: str

        **Examples**

        .. code-block:: python

            promoted_asset_id = client.spaces.promote(asset_id, source_project_id=project_id, target_space_id=space_id)
            promoted_model_id = client.spaces.promote(model_id, source_project_id=project_id, target_space_id=space_id)
            promoted_function_id = client.spaces.promote(function_id, source_project_id=project_id, target_space_id=space_id)
            promoted_data_asset_id = client.spaces.promote(data_asset_id, source_project_id=project_id, target_space_id=space_id)
            promoted_connection_asset_id = client.spaces.promote(connection_id, source_project_id=project_id, target_space_id=space_id)
        """
        promote_payload = {"spaceId": target_space_id,
                           "projectId": source_project_id,
                           "assetDescription": "Asset promoted by ibm_wml client"}

        if rev_id:
            promote_payload['revisionId'] = rev_id

        promote_href = self._client.service_instance._href_definitions.promote_asset_href(asset_id)
        response = requests.post(
            promote_href,
            headers=self._client._get_headers(),
            json=promote_payload
        )
        promotion_details = self._client.repository._handle_response(200, f'promote asset', response)

        try:
            promoted_asset_id = promotion_details['promotedAsset']['asset_id']
        except KeyError as key_err:
            raise PromotionFailed(source_project_id, target_space_id, promotion_details, reason=key_err)

        return promoted_asset_id




"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import torch
import logging
import time
import copy
from importlib import import_module

import numpy as np
import torch.onnx
from sklearn import metrics
from skorch import NeuralNet
from skorch.callbacks import EpochScoring
from skorch.exceptions import NotInitializedError

from ibmfl.model.fl_model import FLModel
from ibmfl.model.model_update import ModelUpdate
from ibmfl.util import config
from ibmfl.exceptions import LocalTrainingException, ModelException, FLException

logger = logging.getLogger(__name__)


class PytorchFLModel(FLModel):
    """
    Wrapper class for importing a Pytorch based model
    """

    def __init__(self, model_name,
                 model_spec=None,
                 pytorch_module=None,
                 module_init_params=None,
                 **kwargs):
        """
        Create a `PytorchFLModel` instance from a Pytorch model.
        If pytorch_model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: String specifying the type of model e.g., Pytorch_NN
        :type model_name: `str`
        :param model_spec: A dictionary specifying path to saved nn.sequence container
        :type model_spec: 'dict'
        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference \
        :param module_init_params: A dictionary with the values for the model's init \
        arguments. The key for the init parameters must be prefixed with module__ \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a PyTorch model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec, **kwargs)

        self.model_type = 'PyTorch'
        self.model = None
        self.module = None
        self.optimizer = torch.optim.SGD
        self.criterion = torch.nn.NLLLoss
        if pytorch_module is None:
            if model_spec is None or (not isinstance(model_spec, dict)):
                raise ValueError('Initializing model requires '
                                 'a model specification or uninstantiated '
                                 'pytorch model class reference '
                                 'None was provided')
            # In this case we need to recreate the model from model_spec
            self.module = self.load_model_from_spec(model_spec)
        else:
            self.module = pytorch_module
        self.model = self.initialize_model(self.module, self.optimizer,
                                           self.criterion, module_init_params)

        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            if self.num_gpus > torch.cuda.device_count():
                logger.error('Selected number of gpus to use for training exceeds number of available gpus, ' +
                             str(torch.cuda.device_count()) +
                             'Set number of gpus to maximum available on device.')
                self.num_gpus = torch.cuda.device_count()

            device_ids = list(range(self.num_gpus))
            self.model.module_ = torch.nn.DataParallel(self.model.module_, device_ids=device_ids)
            self.model.set_params(device='cuda')
            self.model.module_.to('cuda')

    def initialize_model(self, pytorch_module, optimizer, criterion,
                         module_init_params=None):
        """
        Initializes a pytorch model via skorch library

        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference
        :param optimizer: the optimizer to use
        :type optimizer: pytorch optimizer class
        :param criterion: the loss function to use
        :type criterion: pytorch loss function class
        :param module_init_params: A dictionary with the values for the model's init arguments. \
        The key for the init parameters must be prefixed with module__  \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :return: an initialized skorch model
        :rtype: 'skorch.NeuralNet'
        """
        if module_init_params is None:
            module_init_params = {}
        model = NeuralNet(
            module=pytorch_module,
            optimizer=optimizer,
            criterion=criterion,
            warm_start=True,
            callbacks=[('valid_acc', EpochScoring(self.valid_acc,
                                                  lower_is_better=False,
                                                  use_caching=True,
                                                  name='test_valid_acc'))],
            **module_init_params,
        )
        model.initialize()
        return model

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data, a tuple \
        given in the form (x_train, y_train). otherwise, input compatible with skorch.dataset.Dataset
        :type train_data: `(np.ndarray, ”np.ndarray)`
        :param fit_params: (optional) Dictionary with hyperparameters \
        that will be used to call fit function. \
        Hyperparameter parameters should match pytorch expected values \
        e.g., `epochs`, which specifies the number of epochs to be run. \
        If no `epochs` or `batch_size` are provided, a default value \
        will be used (1 and 128, respectively).
        :type fit_params: `dict`
        :return: None
        """
        # Initialized with default values
        batch_size = 128
        epochs = 1
        lr = 0.01
        optimizer_params = {}
        hyperparams = fit_params.get('hyperparams', {}) if fit_params else {}

        if hyperparams:
            local_params = hyperparams.get('local', {}) or {}
            training_hp = local_params.get('training', {}) or {}
            
            epochs = training_hp.get('epochs', epochs)
            batch_size = training_hp.get('batch_size', batch_size)
            lr = training_hp.get('lr', lr)
            
            if 'optimizer' in local_params:
                optimizer = local_params['optimizer']
                if isinstance(optimizer, str):
                    try:
                        optimizer = getattr(import_module(
                            'torch.optim'), optimizer.split('.')[-1])

                    except Exception as e:
                        optimizer = self.optimizer
                        logger.exception(str(e))
                        print('selected optimizer not found. using default')
                if optimizer != self.optimizer:
                    self.optimizer = optimizer
                    self.model.set_params(optimizer=self.optimizer)

                    if self.use_gpu_for_training and torch.cuda.device_count() > 0:
                        self.model.module_.to(self.model.device)

            if 'criterion' in local_params:
                criterion = local_params['criterion']
                if isinstance(criterion, str):
                    try:
                        criterion = getattr(import_module(
                            'torch.nn'), criterion.split('.')[-1])

                    except Exception as e:
                        criterion = self.criterion
                        logger.exception(str(e))
                        print('selected criterion not found. using default')

                if criterion != self.criterion:
                    self.criterion = criterion
                    self.model.set_params(criterion=self.criterion)

            if 'optimizer_params' in local_params:
                optimizer_params = local_params['optimizer_params']
        self.model.set_params(batch_size=batch_size, lr=lr, **optimizer_params)

        try:
            if type(train_data) is tuple:
                # Extract x_train and y_train, by default,
                # label is stored in the last column
                x = train_data[0]
                y = train_data[1]
                self.model.fit_loop(x, y, epochs=epochs)

            else:
                # otherwise, expect that input is a pytorch Dataset generator
                self.model.fit_loop(train_data, epochs=epochs)

        except Exception as e:
            logger.exception(str(e))
            raise LocalTrainingException(
                'Error occurred while performing model.fit')

    def update_model(self, model_update):
        """
        Update model with provided model_update, where model_update
        should be generated according to `PytorchFLModel.get_model_update()`.

        :param model_update: `ModelUpdate` object that contains the weights \
        that will be used to update the model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if isinstance(model_update, ModelUpdate):
            for p1, p2 in zip(self.get_weights(), model_update.get('weights')):
                p1.data = torch.from_numpy(p2)
                p1.data.requires_grad = True

            if self.use_gpu_for_training and torch.cuda.device_count() > 0:
                self.model.module_.to(self.model.device)
        else:
            raise ValueError('Provided model_update should be of type Model.'
                             'Instead they are:{0}'.format(str(type(model_update))))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        weights = self.get_weights(to_numpy=True)
        update = {'weights': weights}

        return ModelUpdate(**update)

    def predict(self, x):
        """
        Perform prediction for a batch of inputs. Note that for classification
        problems, it returns the resulting probabilities.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`, input compatible with skorch.dataset.Dataset, or pytorch dataloader
        :return: Array of predictions
        :rtype: `np.ndarray`
        """
        if isinstance(x, torch.utils.data.DataLoader):
            return self.predict_generator(x)

        return self.model.predict(x)

    def predict_generator(self, dataloader):
        """
        Performs predictions using a pytorch dataloader

        :param x: A pytorch dataloader with a dataset for predicting
        :type x: 'torch.utils.data.DataLoader'
        :return: Array of predictions
        :rtype: 'np.ndarray'
        """
        predictions = None
        for x_batch, _ in dataloader:
            y_pred = self.model.evaluation_step(x_batch)
            y_pred = y_pred.numpy()
            if predictions is None:
                predictions = y_pred
            else:
                predictions = np.append(predictions, y_pred, 0)

        return predictions

    def evaluate(self, test_dataset, eval_metrics=None, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, y_test) or a pytorch DataLoader
        :type test_dataset: `np.ndarray`
        :param eval_metrics: a list of sklearn.metric class, or a function for evaluating a pytorch dataloader batch
        :type eval_metrics: 'sklearn.metrics' or 'function'
        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test, eval_metrics)

        else:
            return self.evaluate_generator_model(test_dataset, eval_metrics)

    def evaluate_model(self, x, y, eval_metrics=None, **kwargs):
        """
        Evaluates the model given x and y.

        :param x:  Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding labels to x
        :type y: `np.ndarray`
        :param eval_metrics: A list of sklearn.metric class or a functions with the signature \
                        (y_true, y_pred, **kwargs) => float
        :type eval_metrics: `sklearn.metrics` or 'function'
        :return: dictionary of metrics
        :rtype: 'dict'
        """
        if eval_metrics is None or len(eval_metrics) == 0:
            eval_metrics = [metrics.accuracy_score]

        # get names for eval_metrics
        metric_names = [getattr(metric, '__name__', repr(metric))
                        for metric in eval_metrics]

        y_pred = self.predict(x)
        y_pred_exp = np.exp(y_pred)
        y_pred_argmax = np.argmax(y_pred_exp, axis=1)

        try:
            metric_dict = dict(zip(metric_names, eval_metrics))
            # NOTE: Had to replace comprehension with for loop to tackle loss fn case which uses y_pred not argmax
            for metric, fn in metric_dict.items():
                if metric is 'loss':
                    metric_dict[metric] = fn(y, y_pred, self.model)
                else:
                    metric_dict[metric] = fn(y, y_pred_argmax, **kwargs)
            # metric_dict = {metric_name: metric(y, y_pred, **kwargs) for metric_name, metric in
            #                zip(metric_names, eval_metrics)}
            return metric_dict
        except TypeError as exc:
            logger.exception(str(exc))
            raise TypeError("eval_metrics must be an sklearn.eval_metrics class, or a function with the signature "
                            "(y_true, y_pred, **kwargs)'")
        except ValueError as exc:
            logger.exception(exc)
            raise ValueError("arguments not in the correct format for metric")

    def evaluate_generator_model(self, dataloader, eval_metrics=None, **kwargs):
        """
        evaluates the model based on the provided dataloader
        :param dataloader: a pytorch dataloader with test dataset and labels
        :type dataloader: torch.utils.data.DataLoader
        :param eval_metrics: a function for how to evaluate the batched examples and labels
        must have the signature (x_batch, y_batch) => float
        :type eval_metrics: 'function'
        :return: dictionary of metrics
        :rtype: 'dict'
        """
        metric_score = 0
        for x_batch, y_batch in dataloader:
            y_pred = self.model.evaluation_step(x_batch)

            if eval_metrics is None:
                y_pred = np.exp(y_pred)
                y_pred = np.argmax(y_pred, axis=1)
                equals = (y_pred == y_batch)
                metric_score += torch.mean(equals.type(torch.FloatTensor))

            else:
                try:
                    metric_score += eval_metrics(x_batch, y_batch)

                except TypeError as exc:
                    logger.exception(str(exc))
                    raise TypeError("eval_metrics must be a function "
                                    "with the signature (x_batch, y_batch, **kwargs)'")

        metric_dict = {'metric_score': metric_score / len(dataloader)}

        return metric_dict

    def save_model(self, filename=None, optimizer_filename=None, history_filename=None, onnx_export=True):
        """
        Save a model to file in the format specific to the backend framework.

        :param filename: Name of the file that contains the model to be loaded.
        :type filename: `str`
        :param optimizer_filename: Name of the file that contains the optimizer to be loaded.
        :type optimizer_filename: `str`
        :param history_filename: Name of the file that contains the model history to be loaded.
        :type history_filename: `str`
        :return: None
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            if onnx_export:
                filename = '{}_{}.onnx'.format(file, time.time())
            else:
                filename = '{}_{}.pt'.format(file, time.time())

        if onnx_export:
            x = torch.randn(1,1,28,28)
            torch.onnx.export(self.model.module_, x, super().get_model_absolute_path(filename), export_params=True, keep_initializers_as_inputs=True)
        else:
            f_params = super().get_model_absolute_path(filename)
            f_optimizer = None
            f_history = None
            if optimizer_filename is not None:
                f_optimizer = super().get_model_absolute_path(optimizer_filename)
            if history_filename is not None:
                f_history = super().get_model_absolute_path(history_filename)
            self.model.save_params(
                f_params=f_params, f_optimizer=f_optimizer, f_history=f_history)

        return filename

    def load_model(self, pytorch_module, model_filename, optimizer_filename=None, history_filename=None, optimizer=None,
                   module_init_params=None):
        """
        Loads a model from disk given the specified file_name

        :param pytorch_module: uninstantiated pytorch model class
        :type pytorch_module: torch.nn.Module class reference
        :param model_filename: Name of the file that contains the model to be loaded.
        :type model_filename: `str`
        :param optimizer_filename: Name of the file that contains the optimizer to be loaded.
        :type optimizer_filename: `str`
        :param history_filename: Name of the file that contains the model history to be loaded.
        :type history_filename: `str`
        :param optimizer: the optimizer that should be loaded
        :type optimizer: pytorch optimizer class
        :param module_init_params: A dictionary with the values for the model's init arguments. \
        The key for the init parameters must be prefixed with module__ \
        (ex. if the parameter name is hidden_size, then key must be module__hidden_size)
        :type module_init_params: 'dict'
        :return: None
        """
        f_params = model_filename
        f_optimizer = None
        f_history = None
        if optimizer is not None and optimizer_filename is not None:
            self.optimizer = optimizer
            f_optimizer = optimizer_filename
        if history_filename is not None:
            f_history = history_filename
        model = self.initialize_model(pytorch_module, self.optimizer, self.criterion,
                                      module_init_params=module_init_params)
        model.load_params(f_params=f_params,
                          f_optimizer=f_optimizer, f_history=f_history)

        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            if self.num_gpus > torch.cuda.device_count():
                logger.error('Selected number of gpus to use for training exceeds number of available gpus, ' +
                             str(torch.cuda.device_count()) +
                             'Set number of gpus to maximum available on device.')
                self.num_gpus = torch.cuda.device_count()
            device_ids = list(range(self.num_gpus))
            model.module_ = torch.nn.DataParallel(model.module_, device_ids=device_ids)
            model.set_params(device='cuda')
            model.module_.to('cuda')

        self.model = model

    def get_weights(self, to_numpy=False):
        """
        Returns the weights of the model

        :param to_numpy; Determines whether the weights should be returned as numpy array, or tensor
        :type to_numpy: `boolean`
        :return: list of model weights
        """
        if self.use_gpu_for_training and torch.cuda.device_count() > 0:
            module = copy.deepcopy(self.model.module_).cpu()

        else:
            module = self.model.module_
        if to_numpy:
            return self.parameters_to_numpy(module.parameters())
        return list(module.parameters())

    def parameters_to_numpy(self, params):
        """
        Transforms parameter tensors to numpy arrays

        :param params: The parameter tensor to be transformed
        :return: numpy array of parameters 
        """
        np_params = []
        for layer in params:
            np_params.append(layer.detach().numpy())
        return np_params

    def load_model_from_spec(self, model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains one item: model_spec['model_definition'], which has a
        pointer to the file where an nn.sequence container is saved

        :return: model
        :rtype: `nn.sequence`
        """

        model_file = model_spec['model_definition']
        model_absolute_path = config.get_absolute_path(model_file)
        model = torch.load(model_absolute_path)
        return model

    def get_gradient(self, train_data):
        """
        Returns the gradients for each layer in the model

        :return: gradients
        :rtype: `list`. numpy array list of model's gradients
        """

        x = train_data[0]
        y = train_data[1]
        gradients = []
        y_infer = self.model.module_(torch.from_numpy(x))
        loss = self.model.criterion_(y_infer, torch.from_numpy(y))
        loss.backward()

        for layer in self.model.module_.parameters():
            gradients.append(layer.grad.numpy())

        return gradients

    @staticmethod
    def valid_acc(net, x, y):
        """
        Callback scoring function that returns the validation accuracy during training.
        :param net: The model that will be used
        :param x: the validation data set
        :param y: the training targets
        :return: the accuracy of the validation training pass
        """
        y_pred = net.predict(x)
        y_pred = np.exp(y_pred)
        y_pred = np.argmax(y_pred, axis=1)
        return metrics.accuracy_score(y, y_pred)

    @staticmethod
    def loss(y, y_pred, net):
        loss = net.get_loss(torch.from_numpy(y_pred), torch.from_numpy(y)).numpy()

        return loss

    @staticmethod
    def __expand_linear_layer__(net, layer_name, new_dimensions, layer_idx):
        """
        Expands a linear layer of a pytorch module
        :param net: an instantiated pytorch module
        :type net: `torch.nn.Module`
        :param layer_name: layer variable name which needs to be expanded
        :type layer_name: `str`
        :param new_dimensions: new dimensions
        :type new_dimensions: `list`
        :param layer_idx: layer idx which needs to be expanded. this corresponds with
        the index of `new_dimensions`
        :type layer_idx: `int`
        :return: None
        """

        original_lin_layer = getattr(net, layer_name)

        if not isinstance(original_lin_layer, torch.nn.Linear):
            raise FLException('Received a non-linear layer to expand '
                              'whereas the method expects a linear layer')

        new_ip_dim = original_lin_layer.in_features if layer_idx == 0 else new_dimensions[
            layer_idx-1]
        new_op_dim = new_dimensions[layer_idx]
        bias = original_lin_layer.bias is not None

        new_lin_layer = type(original_lin_layer)(
            in_features=new_ip_dim, out_features=new_op_dim, bias=bias
        )

        setattr(net, layer_name, new_lin_layer)

    def expand_model_by_layer_name(self, new_dimension, layer_name="dense"):
        """
        Expands the current PyTorch models layers with the provided dimensions
        :param new_dimension: new dimensions of the particular `layer_name`
        :type new_dimension: `list`
        :param layer_name: layer name which needs to be expanded
        :type layer_name: `str`
        :return: None
        """

        if new_dimension is None:
            raise FLException('No information is provided for '
                              'the new expanded model. '
                              'Please provide the new dimension of '
                              'the resulting expanded model')

        layer_maps = {
            'dense': {
                'class': torch.nn.Linear,
                'expansion_fn': self.__expand_linear_layer__
            }
        }

        layer_cls = layer_maps[layer_name]['class']
        layer_exp_fn = layer_maps[layer_name]['expansion_fn']
        net = self.model.module_    # get instantiated module
        idx = 0

        for layer_varname, layer in net.named_modules():
            # skip layer if not instance of layer desired to be expanded
            if not isinstance(layer, layer_cls):
                continue

            layer_exp_fn(net, layer_varname, new_dimension, idx)
            idx += 1

        self.model.initialize_optimizer()

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not.
        In particular, it calls `skorch.utils.check_is_fitted` to checks
        whether the net is initialized.
        If it has, return True; otherwise return false.

        :return: res
        :rtype: `bool`
        """
        try:
            self.model.check_is_fitted()
        except NotInitializedError:
            return False
        return True

    def get_loss(self, dataset):
        """
        Return the resulting loss computed based on the provided dataset.

        :param train_data: Training data, a tuple \
        given in the form (x_train, y_train). otherwise, input compatible with skorch.dataset.Dataset
        :type train_data: `(np.ndarray, ”np.ndarray)`
        :return: The resulting loss.
        :rtype: `float`
        """
        
        metrics_list = [PytorchFLModel.loss]
        orig_loss_value = self.evaluate(dataset, eval_metrics=metrics_list)[
                    'loss']
        
        return orig_loss_value



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
from __future__ import print_function
import joblib
import logging

from sklearn.base import is_classifier
from sklearn.experimental import enable_hist_gradient_boosting
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.ensemble._hist_gradient_boosting.predictor import TreePredictor
from sklearn.ensemble._hist_gradient_boosting.common import PREDICTOR_RECORD_DTYPE

from sklearn.ensemble._hist_gradient_boosting.loss import BinaryCrossEntropy, \
    CategoricalCrossEntropy, LeastSquares

logger = logging.getLogger(__name__)


def export_sklearn(model, full_path=None):
    """
    Auxillary function to first convert the FL XGBoost Model to Scikit Learn's
    Histogram Gradient Boosting model object, then persist the object as a
    pickled model object. This enables standalone use of the trained model
    without dependency to FL's libraries.

    Parameters/Features that are Not Supported or Transferred:
    - monotonic_cst (in v0.21)
    - warm_start
    - early_stopping
    - scoring
    - validation_fraction
    - n_iter_no_change
    - tol

    :param model: The FL XGBoost model object to convert from.
    :type  model: `XGBFLModel`
    :param full_path: The full absolute path of where the model will persist.
    :type  full_path: `str`
    :return: `None`
    """
    # Initialize Model Object (Parameter Transfer from FL Model to Sklearn)
    if model.loss == 'least_squares':
        export_model = HistGradientBoostingRegressor(
            loss=model.loss,
            learning_rate=model.learning_rate,
            max_iter=model.max_iter,
            max_leaf_nodes=model.max_leaf_nodes,
            max_depth=model.max_depth,
            min_samples_leaf=model.min_samples_leaf,
            l2_regularization=model.l2_regularization,
            max_bins=model.max_bins,
            verbose=model.verbose,
            random_state=model.random_state)
    elif model.loss == 'categorical_crossentropy' or \
            model.loss == 'binary_crossentropy' or \
            model.loss == 'auto':
        export_model = HistGradientBoostingClassifier(
            loss=model.loss,
            learning_rate=model.learning_rate,
            max_iter=model.max_iter,
            max_leaf_nodes=model.max_leaf_nodes,
            max_depth=model.max_depth,
            min_samples_leaf=model.min_samples_leaf,
            l2_regularization=model.l2_regularization,
            max_bins=model.max_bins,
            verbose=model.verbose,
            random_state=model.random_state)

    # Attribute Transfer
    export_model._baseline_prediction = model._baseline_prediction
    export_model.n_trees_per_iteration_ = model.n_trees
    export_model.n_features_ = model.n_features_

    # Model Predictor Object (Deep Object Rebuild)
    export_model._predictors = []
    for i, pred in enumerate(model._predictors):
        export_model._predictors.append([])
        for j, p in enumerate(pred):
            nodes = model._predictors[i][j].nodes.astype(
                PREDICTOR_RECORD_DTYPE).copy()
            export_model._predictors[i].append(TreePredictor(nodes))

    # Model Loss Function
    if hasattr(model, 'classes_'):
        if model.loss == 'binary_crossentropy' or len(model.classes_) == 2:
            export_model.loss_ = BinaryCrossEntropy(None)
        elif model.loss == 'categorical_crossentropy':
            export_model.loss_ = CategoricalCrossEntropy(None)
    elif model.loss == 'least_squares':
        export_model.loss_ = LeastSquares(None)

    # Target Class Encoding (For Classification Models)
    if hasattr(model, 'classes_'):
        export_model.classes_ = model.classes_

    # Perform Model Pickling
    with open(full_path, 'wb') as f:
        joblib.dump(export_model, f)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import pkg_resources
from ibm_watson_machine_learning.libs.repo.util.base_singleton import BaseSingleton
from ibm_watson_machine_learning.libs.repo.base_constants import *
import subprocess
import json
class LibraryChecker(BaseSingleton):
    def __init__(self):
        self.supported_libs = [PYSPARK,
                               SCIKIT,
                               PANDAS,
                               XGBOOST,
                               MLPIPELINE,
                               IBMSPARKPIPELINE,
                               TENSORFLOW]
        self.installed_libs = {PYSPARK: False,
                               SCIKIT: False,
                               PANDAS: False,
                               XGBOOST: False,
                               MLPIPELINE: False,
                               TENSORFLOW: False,
                               IBMSPARKPIPELINE: False}
        atleast_one_lib_installed = self._check_if_lib_installed(self.supported_libs)

        if not atleast_one_lib_installed:
            supported_lib_str = self.supported_libs[0]
            lib_num = len(self.supported_libs)
            for i in range(1, lib_num-1):
                supported_lib_str += ', ' + self.supported_libs[i]
            supported_lib_str += ' and ' + self.supported_libs[lib_num-1]
            raise ImportError("The system lacks installations of " + supported_lib_str +
                              ". At least one of the libraries is required for the repository-client to be used")

    def _check_if_lib_installed(self, lib_names):
        import sys
        atleast_one_lib_installed = False

        # Using try except method as these libraries aren't present in the
        # PyPI repository for pip
        try:
            import pyspark
            self.installed_libs[PYSPARK] = True
            atleast_one_lib_installed = True
        except ImportError:
            pass
        try:
            import mlpipelinepy
            self.installed_libs[MLPIPELINE] = True
            atleast_one_lib_installed = True
        except ImportError:
            pass
        try:
            import pipeline
            self.installed_libs[IBMSPARKPIPELINE] = True
            atleast_one_lib_installed = True
        except ImportError:
            pass

        installed_pkgs = pkg_resources.working_set
        pkg_list = sorted(["%s==%s" % (i.key, i.version) for i in installed_pkgs])
        for name in lib_names:
            for pkg in pkg_list:
                if name in str(pkg):
                    self.installed_libs[name] = True
                    atleast_one_lib_installed = True
                if 'scikit-learn' == str(pkg):
                    import sklearn
                    import sys
                    if sklearn.__version__ == '0.23.0' and sys.version_info <= (3, 7):
                        raise Exception(' Scikit learn version 0.23.0 is not supported, Please downgrade scikit version to a lower version and re-try. ')

        return atleast_one_lib_installed

    def check_lib(self, lib_name):
        lib_display_names={PYSPARK: DISPLAY_PYSPARK,
                           SCIKIT: DISPLAY_SCIKIT,
                           PANDAS: DISPLAY_PANDAS,
                           XGBOOST: DISPLAY_XGBOOST,
                           MLPIPELINE: DISPLAY_MLPIPELINE,
                           TENSORFLOW: DISPLAY_TENSORFLOW,
                           IBMSPARKPIPELINE: DISPLAY_IBMSPARKPIPELINE}
        if not self.installed_libs[lib_name]:
            raise NameError('{} Library is not installed. Please install it and execute the command'.
                            format(lib_display_names[lib_name]))


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import abc

class CryptoHe(abc.ABC):
    """
    This class defines an interface for HE keys generation functions. 
    """

    @abc.abstractmethod
    def __init__(self, config = None):
        return

    @abc.abstractmethod
    def generate_keys(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_public_key(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_private_key(self):
        raise NotImplementedError



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
import json
from ibm_watson_machine_learning.utils import INSTANCE_DETAILS_TYPE, print_text_header_h2
from ibm_watson_machine_learning.wml_client_error import WMLClientError, WrongMetaProps
from ibm_watson_machine_learning.wml_resource import WMLResource
from .metanames import FactsheetsMetaNames


class Factsheets(WMLResource):
    """Link WML Model to Model Entry."""

    cloud_platform_spaces = False
    icp_platform_spaces = False

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        if not client.ICP and not client.CLOUD_PLATFORM_SPACES and not client.ICP_PLATFORM_SPACES:
            Factsheets._validate_type(client.service_instance.details, u'instance_details', dict, True)
            Factsheets._validate_type_of_details(client.service_instance.details, INSTANCE_DETAILS_TYPE)
        self._ICP = client.ICP
        self.ConfigurationMetaNames = FactsheetsMetaNames()

        if client.CLOUD_PLATFORM_SPACES:
            Factsheets.cloud_platform_spaces = True

        if client.ICP_PLATFORM_SPACES:
            Factsheets.icp_platform_spaces = True

    def register_model_entry(self, model_id, meta_props, catalog_id=None):
        """Link WML Model to Model Entry

        :param model_id: published model/asset ID
        :type model_id: str

        :param meta_props: metaprops, to see the available list of metanames use:

            .. code-block:: python

                client.factsheets.ConfigurationMetaNames.get()

        :type meta_props: dict

        :param catalog_id: catalog ID where you want to register model
        :type catalog_id: str, optional

        :return: metadata of the registration
        :rtype: dict

        **Example**

        .. code-block:: python

            meta_props = {
                wml_client.factsheets.ConfigurationMetaNames.ASSET_ID: '83a53931-a8c0-4c2f-8319-c793155e7517'}

            registration_details = client.factsheets.register_model_entry(model_id, catalog_id, meta_props)

        or

        .. code-block:: python

            meta_props = {
                wml_client.factsheets.ConfigurationMetaNames.NAME: "New model entry",
                wml_client.factsheets.ConfigurationMetaNames.DESCRIPTION: "New model entry"}

            registration_details = client.factsheets.register_model_entry(model_id, meta_props)

        """
        Factsheets._validate_type(model_id, u'model_id', str, True)
        Factsheets._validate_type(catalog_id, u'catalog_id', str, False)
        metaProps = self.ConfigurationMetaNames._generate_resource_metadata(meta_props)

        params = self._client._params()

        if catalog_id is not None:
            params['catalog_id'] = catalog_id
            if 'project_id' in params:
                del params['project_id']
            elif 'space_id' in params:
                del params['space_id']

        name_in = self.ConfigurationMetaNames.NAME in metaProps
        description_in = self.ConfigurationMetaNames.DESCRIPTION in metaProps
        asset_id_in = self.ConfigurationMetaNames.ASSET_ID in metaProps

        # check for metaprops correctness
        reason = "Please provide either NAME and DESCRIPTION or ASSET_ID"
        if name_in and description_in:
            if asset_id_in:
                raise WrongMetaProps(reason=reason)

        elif asset_id_in:
            if name_in or description_in:
                raise WrongMetaProps(reason=reason)

        else:
            raise WrongMetaProps(reason=reason)

        url = self._client.service_instance._href_definitions.get_wkc_model_register_href(model_id)

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            response = requests.post(
                url,
                json=metaProps,
                params=params,  # version is mandatory
                headers=self._client._get_headers())
        else:
            response = requests.post(
                url,
                json=metaProps,
                headers=self._client._get_headers())

        if response.status_code == 200:
            print_text_header_h2(f"Successfully finished linking WML Model '{model_id}' to Model Entry.")

        else:
            error_msg = u'WML Model registration failed'
            reason = response.text
            print(reason)
            print_text_header_h2(error_msg)
            raise WMLClientError(error_msg + '. Error: ' + str(response.status_code) + '. ' + reason)

        return response.json()

    def list_model_entries(self, catalog_id=None):
        """Returns all WKC Model Entry assets for a catalog.

        :param catalog_id: catalog ID where you want to register model, if None list from all catalogs
        :type catalog_id: str, optional

        :return: all WKC Model Entry assets for a catalog
        :rtype: dict

        **Example**

        .. code-block:: python

            model_entries = client.factsheets.list_model_entries(catalog_id)

        """
        if catalog_id is not None:
            Factsheets._validate_type(catalog_id, u'catalog_id', str, True)
            url = self._client.service_instance._href_definitions.get_wkc_model_list_from_catalog_href(catalog_id)

        else:
            url = self._client.service_instance._href_definitions.get_wkc_model_list_all_href()

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            response = requests.get(
                url,
                params=self._client._params(),  # version is mandatory
                headers=self._client._get_headers())
        else:
            response = requests.get(
                url,
                headers=self._client._get_headers())

        if response.status_code == 200:
            return response.json()

        else:
            error_msg = u'WKC Models listing failed'
            reason = response.text
            print(reason)
            print_text_header_h2(error_msg)
            raise WMLClientError(error_msg + '. Error: ' + str(response.status_code) + '. ' + reason)

    def unregister_model_entry(self, asset_id, catalog_id: str = None):
        """Unregister WKC Model Entry

        :param asset_id:  WKC model entry id
        :type asset_id: str

        :param catalog_id: catalog ID where asset is stored, when not provided,
            default client space or project will be taken
        :type catalog_id: str, optional

        **Example**

        .. code-block:: python

            model_entries = client.factsheets.unregister_model_entry(asset_id='83a53931-a8c0-4c2f-8319-c793155e7517',
                                                                     catalog_id='34553931-a8c0-4c2f-8319-c793155e7517')

        or

        .. code-block:: python

            client.set.default_space('98f53931-a8c0-4c2f-8319-c793155e7517')
            model_entries = client.factsheets.unregister_model_entry(asset_id='83a53931-a8c0-4c2f-8319-c793155e7517')

        """
        Factsheets._validate_type(asset_id, u'asset_id', str, True)
        Factsheets._validate_type(catalog_id, u'catalog_id', str, False)
        url = self._client.service_instance._href_definitions.get_wkc_model_delete_href(asset_id)

        params = self._client._params()
        if catalog_id is not None:
            params['catalog_id'] = catalog_id

            if 'space_id' in str(params):
                del params['space_id']

            elif 'project_id' in str(params):
                del params['project_id']

        response = requests.delete(
            url,
            params=params,  # version is mandatory
            headers=self._client._get_headers())

        if response.status_code == 204:
            print_text_header_h2(f"Successfully finished unregistering WKC Model '{asset_id}' Entry.")

        else:
            error_msg = u'WKC Model Entry unregistering failed'
            reason = response.text
            print(reason)
            print_text_header_h2(error_msg)
            raise WMLClientError(error_msg + '. Error: ' + str(response.status_code) + '. ' + reason)



import abc
import numpy as np
import logging
from pickle import loads, dumps
from ibmfl.exceptions import ModelUpdateException

logger = logging.getLogger(__name__)


class ModelUpdate():
    """
    Class to hold model update dictionary. Dictionary can only be accessed
    using inbuild methods like `get` and `add`
    """

    def __init__(self, **kwargs):
        """
        Initialize the dictionary and add updates.
        :param kwargs: Dictionary of model-update specific arguments.
        :type kwargs: `dict`
        """
        self.__updates = {}
        for key, value in kwargs.items():
            self.add(key, value)

    def add(self, key, value):
        """
        Add update `value` for `key`

        :param key: Identifier which represents the update
        :type key: `str`
        :param value: Content of the update
        :type value: any ds/object which can be pickled
        """
        try:
            self.__updates[key] = dumps(value)
        except Exception as ex:
            logger.exception(ex)

            logger.exception("Error occured while adding a update.\
                                Make sure model update data structure is picklable")

            raise ModelUpdateException("Error updating model update")

    def get(self, key):
        """
        Get update value for given key from model update dictionary

        :param key: Identifier which represents the update
        :type key: `str`
        :return: content of the update
        :rtype: any ds/object after its unpickled
        """
        if key not in self.__updates:
            logger.error("Key "+key+" not found in model updates")
            raise ModelUpdateException(
                "Invalid key was requested from model update")

        ret_val = loads(self.__updates[key])
        return ret_val

    def exist_key(self, key):
        """
        Check if a specified key is used in model update dictionary

        :param key: Identifier which represents the update
        :type key: `str`
        :return: check result of existing the key
        :rtype: True or False
        """
        return True if key in self.__updates else False



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = ["ExperimentIterableDataset"]

import os
import pandas as pd

from re import findall
from typing import Union, Optional
from warnings import warn
from ibm_watson_machine_learning.helpers.connections.local import LocalBatchReader
from ibm_watson_machine_learning.utils.autoai.enums import SamplingTypes

# Note: try to import torch lib if available, this fallback is done based on
# torch dependency removal request
try:
    from torch.utils.data import IterableDataset

except ImportError:
    IterableDataset = object
# --- end note

DEFAULT_SAMPLE_SIZE_LIMIT = 1073741824  # 1GB in Bytes is verified later by _set_sample_size_limit
DEFAULT_REDUCED_SAMPLE_SIZE_LIMIT = 104857600  # 100MB in bytes
DEFAULT_SAMPLING_TYPE = SamplingTypes.FIRST_VALUES


class ExperimentIterableDataset(IterableDataset):
    """
        This dataset is intended to be an Iterable stream from Flight Service.
        It should iterate over flight logical batches and manages by Connection class
        how batches are downloaded and created. It should take into consideration only 2 batches at a time.
        If we have 2 batches already downloaded, it should block further download
        and wait for first batch to be consumed.

        Example
        -------
        >>> experiment_metadata = {
        >>>     "prediction_column": 'species',
        >>>     "prediction_type": "classification",
        >>>     "project_id": os.environ.get('PROJECT_ID'),
        >>>     'wml_credentials': wml_credentials
        >>> }

        >>> connection = DataConnection(data_asset_id='5d99c11a-2060-4ef6-83d5-dc593c6455e2')


        >>># default sampling - read first 1GB of data
        >>> iterable_dataset = ExperimentIterableDataset(connection=connection,
        >>>                                              enable_sampling=True,
        >>>                                              sampling_type='first_n_records',
        >>>                                              sample_size_limit = 1GB,
        >>>                                              experiment_metadata=experiment_metadata)

        >>># read all data records in batches / no subsampling
        >>> iterable_dataset = ExperimentIterableDataset(connection=connection,
        >>>                                              enable_sampling=False,
        >>>                                               experiment_metadata=experiment_metadata)

        >>># stratified/random sampling
        >>> iterable_dataset = ExperimentIterableDataset(connection=connection,
        >>>                                              enable_sampling=True,
        >>>                                              sampling_type='stratified',
        >>>                                              sample_size_limit = 1GB,
        >>>                                              experiment_metadata=experiment_metadata)

    """

    connection: Union["FlightConnection", LocalBatchReader] = None

    def __init__(
        self,
        connection: "DataConnection",
        experiment_metadata: dict = None,
        enable_sampling: bool = True,
        sample_size_limit: int = DEFAULT_SAMPLE_SIZE_LIMIT,
        sampling_type: str = DEFAULT_SAMPLING_TYPE,
        binary_data: bool = False,
        number_of_batch_rows: int = None,
        stop_after_first_batch: bool = False,
        total_size_limit=DEFAULT_SAMPLE_SIZE_LIMIT,
        total_nrows_limit=None,
        total_percentage_limit=1.0,
        **kwargs,
    ):
        super().__init__()
        self.enable_sampling = enable_sampling
        self.sample_size_limit = sample_size_limit
        self.experiment_metadata = experiment_metadata
        self._wml_client = getattr(connection, "_wml_client", None)
        if self._wml_client is None:
            self._wml_client = kwargs.get("_wml_client", None)
        self.binary_data = binary_data
        self.sampling_type = sampling_type
        self.read_to_file = kwargs.get('read_to_file')
        self.authorized = self._wml_check_authorization()
        self._set_size_limit(total_size_limit)
        self.total_nrows_limit = total_nrows_limit
        self.total_percentage_limit = total_percentage_limit

        # Note: convert to dictionary if we have object from WML client
        if not isinstance(connection, dict):
            dict_connection = connection._to_dict()

        else:
            dict_connection = connection
        # --- end note

        # Note: backward compatibility after sampling refactoring #27255
        if kwargs.get('with_sampling') or kwargs.get('normal_read'):
            warn("The parameters with_sampling and normal_read in ExperimentIterableDataset are deprecated. "
                 "Use enable_sampling and sampling_type instead.")
            if kwargs.get('normal_read'):
                self.enable_sampling = False
            if kwargs.get('with_sampling'):
                from ibm_watson_machine_learning.utils.autoai.enums import PredictionType
                self.enable_sampling = True
                if self.experiment_metadata.get("prediction_type") in [PredictionType.REGRESSION]:
                    self.sampling_type = SamplingTypes.RANDOM
                elif self.experiment_metadata.get("prediction_type") in [PredictionType.CLASSIFICATION,
                                                                         PredictionType.BINARY,
                                                                         PredictionType.MULTICLASS]:
                    self.sampling_type = SamplingTypes.STRATIFIED
        # --- end note

        # if number_of_batch_rows is provided, batch_size does not matter anymore
        if self.authorized:
            is_cos_asset = bool(kwargs.get("flight_parameters", {}).get('datasource_type', {}).get('entity', {}).get('name', '') == 'bluemixcloudobjectstorage')
            # first used headers from experiment metadata if they were set.
            headers_ = None
            if self.experiment_metadata.get("headers"):
                headers_ = self.experiment_metadata.get("headers")
            elif self._wml_client is not None:
                headers_ = self._wml_client._get_headers()

            from ibm_watson_machine_learning.helpers.connections.flight_service import FlightConnection

            self.connection = FlightConnection(
                headers=headers_,
                sampling_type=self.sampling_type,
                label=self.experiment_metadata.get("prediction_column"),
                learning_type=self.experiment_metadata.get("prediction_type"),
                params=self.experiment_metadata,
                project_id=self.experiment_metadata.get("project_id", getattr(self._wml_client, 'default_project_id', None)),
                space_id=self.experiment_metadata.get("space_id", getattr(self._wml_client, 'default_space_id', None)),
                asset_id=None if is_cos_asset else dict_connection.get("location", {}).get("id"), # do not pass asset id for data assets located on COS
                connection_id=dict_connection.get("connection", {}).get("id"),
                data_location=dict_connection,
                data_batch_size_limit=self.sample_size_limit,
                flight_parameters=kwargs.get("flight_parameters", {}),
                extra_interaction_properties=kwargs.get("extra_interaction_properties",{}),
                fallback_to_one_connection=kwargs.get(
                    "fallback_to_one_connection", True
                ),
                number_of_batch_rows=number_of_batch_rows,
                stop_after_first_batch=stop_after_first_batch,
                _wml_client=self._wml_client,
                return_subsampling_stats=kwargs.get("_return_subsampling_stats", False),
                total_size_limit=self.total_size_limit,
                total_nrows_limit=self.total_nrows_limit,
                total_percentage_limit=self.total_percentage_limit,
            )

        else:

            if (
                dict_connection.get("type") == "fs"
                and "location" in dict_connection
                and "path" in dict_connection["location"]
            ):
                self.connection = LocalBatchReader(
                    file_path=dict_connection["location"]["path"], batch_size=sample_size_limit
                )

            else:
                raise NotImplementedError(
                    "For local data read please use 'fs' (file system) connection type."
                    "To use remote data read please provide 'experiment_metadata'."
                )

    def _wml_check_authorization(self) -> bool:
        """
        Check if we can authorize with WML.
        If the connection have wml_client initialized use it as an attribute.
        Otherwise, credentials should be provided in the experiment_metadata dictionary.
        If the client is properly initialized True will be returned.
        """
        if self._wml_client is not None:
            return True

        if self.experiment_metadata is None:
            return False

        if self.experiment_metadata.get("wml_credentials") is not None:
            from ibm_watson_machine_learning import APIClient

            self._wml_client = APIClient(
                wml_credentials=self.experiment_metadata["wml_credentials"]
            )
            return True

        elif self.experiment_metadata.get("headers") is not None:
            return True

        else:
            return False

    def _set_size_limit(self, size_limit):
        """If non-default value of total_size_limit was not passed,
        set Sample Size Limit based on T-Shirt size if code is run on training pod:
        For memory < 32 (T-Shirts: XS,S, M) default is 100MB,
        For memory = 32 (T-Shirt L) default is 0.7GB,
        For memory > 32 (T-Shirt XL) or runs outside pod default is 1GB.
        """
        if size_limit == DEFAULT_SAMPLE_SIZE_LIMIT:
            size_mem = int(findall(r'[0-9]+', os.environ.get('MEM', '32'))[0])
            if size_mem > 32:
                self.total_size_limit = DEFAULT_SAMPLE_SIZE_LIMIT             # 1GB in bytes
            elif size_mem == 32:
                self.total_size_limit = int(0.7 * DEFAULT_SAMPLE_SIZE_LIMIT)  # 0.7GB in bytes
            else:
                self.total_size_limit = DEFAULT_REDUCED_SAMPLE_SIZE_LIMIT     # 100MB in bytes
        else:
            self.total_size_limit = size_limit

    def write(
        self, data: Optional[pd.DataFrame] = None, file_path: Optional[str] = None
    ) -> None:
        """
        Writes data into data source connection.

        Parameters
        ----------
        data: pandas DataFrame, optional
            Structured data to be saved in dat source connection. (Either 'data' or 'file_path' need to be provided)

        file_path: str, optional
            Path to the local file to be saved in source data connection (binary transfer).
            (Either 'data' or 'file_path' need to be provided)
        """
        if (data is None and file_path is None) or (
            data is not None and file_path is not None
        ):
            raise ValueError("Either 'data' or 'file_path' need to be provided.")

        if data is not None and not isinstance(data, pd.DataFrame):
            raise TypeError(
                f"'data' need to be a pandas DataFrame, you provided: '{type(data)}'."
            )

        if file_path is not None and not isinstance(file_path, str):
            raise TypeError(
                f"'file_path' need to be a string, you provided: '{type(file_path)}'."
            )

        if data is not None:
            self.connection.write_data(data)

        else:
            self.connection.write_binary_data(file_path)

    def __iter__(self):
        """Iterate over Flight Dataset."""
        if self.authorized:
            if self.enable_sampling:
                if self.sampling_type == SamplingTypes.FIRST_VALUES:
                    return self.connection.iterable_read()
                else:
                    self.connection.enable_subsampling = True
                    return self.connection.iterable_read()
            else:
                if self.binary_data:
                    return self.connection.read_binary_data(read_to_file=self.read_to_file)

                else:
                    self.total_size_limit = None
                    return self.connection.iterable_read()
        else:
            return (batch for batch in self.connection)



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class MlAssetsUploadContentOutputMetadata(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self):
        """
        MlAssetsUploadContentOutputMetadata - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'guid': 'str',
            'created_at': 'datetime',
            'modified_at': 'datetime',
            'url': 'str'
        }

        self.attribute_map = {
            'guid': 'guid',
            'created_at': 'created_at',
            'modified_at': 'modified_at',
            'url': 'url'
        }

        self._guid = None
        self._created_at = None
        self._modified_at = None
        self._url = None

    @property
    def guid(self):
        """
        Gets the guid of this MlAssetsUploadContentOutputMetadata.


        :return: The guid of this MlAssetsUploadContentOutputMetadata.
        :rtype: str
        """
        return self._guid

    @guid.setter
    def guid(self, guid):
        """
        Sets the guid of this MlAssetsUploadContentOutputMetadata.


        :param guid: The guid of this MlAssetsUploadContentOutputMetadata.
        :type: str
        """
        self._guid = guid

    @property
    def created_at(self):
        """
        Gets the created_at of this MlAssetsUploadContentOutputMetadata.


        :return: The created_at of this MlAssetsUploadContentOutputMetadata.
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """
        Sets the created_at of this MlAssetsUploadContentOutputMetadata.


        :param created_at: The created_at of this MlAssetsUploadContentOutputMetadata.
        :type: datetime
        """
        self._created_at = created_at

    @property
    def modified_at(self):
        """
        Gets the modified_at of this MlAssetsUploadContentOutputMetadata.


        :return: The modified_at of this MlAssetsUploadContentOutputMetadata.
        :rtype: datetime
        """
        return self._modified_at

    @modified_at.setter
    def modified_at(self, modified_at):
        """
        Sets the modified_at of this MlAssetsUploadContentOutputMetadata.


        :param modified_at: The modified_at of this MlAssetsUploadContentOutputMetadata.
        :type: datetime
        """
        self._modified_at = modified_at

    @property
    def url(self):
        """
        Gets the url of this MlAssetsUploadContentOutputMetadata.


        :return: The url of this MlAssetsUploadContentOutputMetadata.
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """
        Sets the url of this MlAssetsUploadContentOutputMetadata.


        :param url: The url of this MlAssetsUploadContentOutputMetadata.
        :type: str
        """
        self._url = url

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import abc
from ibmfl.crypto.crypto_exceptions import *

class CryptoCert:
    """
    This class defines an interface for certificate verification functions. 
    """

    @abc.abstractmethod
    def __init__(self, ca_cert_file_path: str, my_cert_file_path: str, **kwargs):
        return

    @abc.abstractmethod
    def verify_cert_signature(self, certificate):
        raise NotImplementedError

    def verify_certs(self, certificates):
        ret = {}
        for id, cert in certificates.items():
            ver, pbkey = self.verify_cert_signature(cert)
            if not ver:
                raise KeyDistributionVerificationException("Invalid certificate=" + repr(cert))
            ret[id] = pbkey
        return ret

    @abc.abstractmethod
    def get_my_cert(self, ret_type: str):
        raise NotImplementedError



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.function_artifact_loader  import FunctionArtifactLoader


class FunctionLoader(FunctionArtifactLoader):
    """
        Returns  Generic function instance associated with this function artifact.

        :return: function
        :rtype:
        """

    def function_instance(self):
        """
         :return: returns function path
         """
        return self.load()

    def download_function(self,path):
        """
         :return: returns function path
         """
        return self.load(path)




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
import sys

__all__ = [
    "WMLClientError",
    "MissingValue",
    "MissingMetaProp",
    "NotUrlNorUID",
    "NoWMLCredentialsProvided",
    "ApiRequestFailure",
    "UnexpectedType",
    "ForbiddenActionForPlan",
    "NoVirtualDeploymentSupportedForICP",
    "MissingArgument",
    "WrongEnvironmentVersion",
    "CannotAutogenerateBedrockUrl",
    "WrongMetaProps",
    "CannotSetProjectOrSpace",
    "ForbiddenActionForGitBasedProject",
    "CannotInstallLibrary",
    "DataStreamError",
    "WrongLocationProperty",
    "WrongFileLocation",
    "EmptyDataSource",
    "SpaceIDandProjectIDCannotBeNone"
]


class WMLClientError(Exception):
    def __init__(self, error_msg, reason=None, logg_messages=True):
        self.error_msg = error_msg
        self.reason = reason
        if logg_messages:
            logging.getLogger(__name__).warning(self.__str__())
            logging.getLogger(__name__).debug(
                str(self.error_msg) + ('\nReason: ' + str(self.reason) if sys.exc_info()[0] is not None else ''))

    def __str__(self):
        return str(self.error_msg) + ('\nReason: ' + str(self.reason) if self.reason is not None else '')


class MissingValue(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, 'No \"' + value_name + '\" provided.', reason)


class MissingMetaProp(MissingValue):
    def __init__(self, name, reason=None):
        WMLClientError.__init__(self, 'Missing meta_prop with name: \'{}\'.'.format(name), reason)


class NotUrlNorUID(WMLClientError, ValueError):
    def __init__(self, value_name, value, reason=None):
        WMLClientError.__init__(self,
                                'Invalid value of \'{}\' - it is not url nor uid: \'{}\''.format(value_name, value),
                                reason)


class NoWMLCredentialsProvided(MissingValue):
    def __init__(self, reason=None):
        MissingValue.__init__(self, 'WML credentials', reason)


class ApiRequestFailure(WMLClientError):
    def __init__(self, error_msg, response, reason = None):
        self.response = response
        if str(response.status_code) == '404' and 'DOCTYPE' in str(response.content):
            raise MissingWMLComponent()

        elif (str(response.status_code) == '400' and
              'Invalid content. You cannot include any tags in the HTTP request.' in str(response.content)):
            WMLClientError.__init__(self, f"Please check if any parameter that you provided include HTTP tag. "
                                          f"If yes, please remove it and try again.", reason=str(response.content))

        else:
            WMLClientError.__init__(self,
                                    '{} ({} {})\nStatus code: {}, body: {}'.format(
                                        error_msg,
                                        response.request.method,
                                        response.request.url,
                                        response.status_code,
                                        response.text if response.apparent_encoding is not None else '[binary content, ' + str(
                                            len(response.content)) + ' bytes]'), reason)


class UnexpectedType(WMLClientError, ValueError):
    def __init__(self, el_name, expected_type, actual_type):
        WMLClientError.__init__(self, 'Unexpected type of \'{}\', expected: {}, actual: \'{}\'.'.format(el_name,
                                                                                                        '\'{}\''.format(
                                                                                                            expected_type) if type(
                                                                                                            expected_type) == type else expected_type,
                                                                                                        actual_type))


class ForbiddenActionForPlan(WMLClientError):
    def __init__(self, operation_name, expected_plans, actual_plan):
        WMLClientError.__init__(self,
                                'Operation \'{}\' is available only for {} plan, while this instance has \'{}\' plan.'.format(
                                    operation_name, ('one of {} as'.format(expected_plans) if len(
                                        expected_plans) > 1 else '\'{}\''.format(expected_plans[0])) if type(
                                        expected_plans) is list else '\'{}\''.format(expected_plans), actual_plan))


class NoVirtualDeploymentSupportedForICP(MissingValue):
    def __init__(self, reason=None):
        MissingValue.__init__(self, 'No Virtual deployment supported for ICP', reason)


class MissingArgument(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Argument: {value_name} missing.", reason)


class WrongEnvironmentVersion(WMLClientError, ValueError):
    def __init__(self, used_version, environment_name, supported_versions):
        WMLClientError.__init__(self, "Version used in credentials not supported in this environment",
                                reason=f"Version {used_version} isn't supported in "
                                       f"{environment_name} environment, "
                                       f"select from {supported_versions}")


class CannotAutogenerateBedrockUrl(WMLClientError, ValueError):
    def __init__(self, e1, e2):
        WMLClientError.__init__(self, "Attempt of generating `bedrock_url` automatically failed. "
                                      "If iamintegration=True, please provide `bedrock_url` in wml_credentials. "
                                      "If iamintegration=False, please validate your credentials.", reason=[e1, e2])


class WrongMetaProps(MissingValue):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, "Wrong metaprops.", reason)


class MissingWMLComponent(WMLClientError):
    def __init__(self):
        WMLClientError.__init__(self, f"Missing WML Component",
                                reason=f"It appears that WML component is not installed on your environment. "
                                       f"Contact your cluster administrator.")


class CannotSetProjectOrSpace(WMLClientError):
    def __init__(self, reason):
        WMLClientError.__init__(self, f"Cannot set Project or Space",
                                reason=reason)


class ForbiddenActionForGitBasedProject(WMLClientError):
    def __init__(self, reason):
        WMLClientError.__init__(self, f"This action is not supported for git based project.",
                                reason=reason)


class CannotInstallLibrary(WMLClientError, ValueError):
    def __init__(self, lib_name: str, reason: str):
        WMLClientError.__init__(self, f"Library '{lib_name}' cannot be installed! Please install it manually.", reason)


class DataStreamError(WMLClientError, ConnectionError):
    def __init__(self, reason):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class WrongLocationProperty(WMLClientError, ConnectionError):
    def __init__(self, reason):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class WrongFileLocation(WMLClientError, ValueError):
    def __init__(self, reason):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)

class EmptyDataSource(WMLClientError, ValueError):
    def __init__(self ):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. "
                                      "Verify if data were saved under data connection and try again.")

class SpaceIDandProjectIDCannotBeNone(WMLClientError, ValueError):
    def __init__(self, reason: str):
        WMLClientError.__init__(self, f"Missing 'space_id' or 'project_id'.", reason)



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class ExperimentInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, tags=None, settings=None, training_references=None, training_data_reference=None, training_results_reference=None):
        """
        ExperimentInputExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'tags': 'list[TagRepository]',
            'settings': 'ExperimentInputSettings',
            'training_references': 'list[TrainingReferenceExperiments]',
            'training_data_reference': 'ConnectionObjectSourceExperiments',
            'training_results_reference': 'ConnectionObjectTargetExperiments'
        }

        self.attribute_map = {
            'tags': 'tags',
            'settings': 'settings',
            'training_references': 'training_references',
            'training_data_reference': 'training_data_reference',
            'training_results_reference': 'training_results_reference'
        }

        self._tags = tags
        self._settings = settings
        self._training_references = training_references
        self._training_data_reference = training_data_reference
        self._training_results_reference = training_results_reference

    @property
    def tags(self):
        """
        Gets the tags of this ExperimentInputExperiments.


        :return: The tags of this ExperimentInputExperiments.
        :rtype: list[TagRepository]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of this ExperimentInputExperiments.


        :param tags: The tags of this ExperimentInputExperiments.
        :type: list[TagRepository]
        """

        self._tags = tags

    @property
    def settings(self):
        """
        Gets the settings of this ExperimentInputExperiments.


        :return: The settings of this ExperimentInputExperiments.
        :rtype: ExperimentInputSettings
        """
        return self._settings

    @settings.setter
    def settings(self, settings):
        """
        Sets the settings of this ExperimentInputExperiments.


        :param settings: The settings of this ExperimentInputExperiments.
        :type: ExperimentInputSettings
        """

        self._settings = settings

    @property
    def training_references(self):
        """
        Gets the training_references of this ExperimentInputExperiments.


        :return: The training_references of this ExperimentInputExperiments.
        :rtype: list[TrainingReferenceExperiments]
        """
        return self._training_references

    @training_references.setter
    def training_references(self, training_references):
        """
        Sets the training_references of this ExperimentInputExperiments.


        :param training_references: The training_references of this ExperimentInputExperiments.
        :type: list[TrainingReferenceExperiments]
        """

        self._training_references = training_references

    @property
    def training_data_reference(self):
        """
        Gets the training_data_reference of this ExperimentInputExperiments.


        :return: The training_data_reference of this ExperimentInputExperiments.
        :rtype: ConnectionObjectSourceExperiments
        """
        return self._training_data_reference

    @training_data_reference.setter
    def training_data_reference(self, training_data_reference):
        """
        Sets the training_data_reference of this ExperimentInputExperiments.


        :param training_data_reference: The training_data_reference of this ExperimentInputExperiments.
        :type: ConnectionObjectSourceExperiments
        """

        self._training_data_reference = training_data_reference

    @property
    def training_results_reference(self):
        """
        Gets the training_results_reference of this ExperimentInputExperiments.
        This is required for DL

        :return: The training_results_reference of this ExperimentInputExperiments.
        :rtype: ConnectionObjectTargetExperiments
        """
        return self._training_results_reference

    @training_results_reference.setter
    def training_results_reference(self, training_results_reference):
        """
        Sets the training_results_reference of this ExperimentInputExperiments.
        This is required for DL

        :param training_results_reference: The training_results_reference of this ExperimentInputExperiments.
        :type: ConnectionObjectTargetExperiments
        """

        self._training_results_reference = training_results_reference

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other
